"""
Tests for dcmstack.dcmstack
"""
from __future__ import absolute_import, print_function

import sys
import warnings
from copy import deepcopy
from glob import glob
from hashlib import sha256
from os import path

import numpy as np
import pytest

from . import test_dir, src_dir

try:
    import pydicom
    import pydicom.dataset
    from pydicom.datadict import keyword_dict, dictionary_VR
    from pydicom.uid import ExplicitVRLittleEndian
except ImportError:
    import dicom as pydicom
    from dicom.datadict import keyword_dict
    from dicom.datadict import dictionaryVR as dictionary_VR
    from dicom.UID import ExplicitVRLittleEndian
import nibabel as nb
from nibabel.orientations import aff2axcodes

import dcmstack

if hasattr(pydicom.dataset, "FileMetaDataset"):
    _def_file_meta = pydicom.dataset.FileMetaDataset()
else:
     _def_file_meta = pydicom.dataset.Dataset()
_def_file_meta.TransferSyntaxUID = ExplicitVRLittleEndian

def_dicom_attrs = {'file_meta' : _def_file_meta,
                   'ImagePositionPatient' : [0.0, 0.0, 0.0],
                   'ImageOrientationPatient' : [1.0, 0.0, 0.0, 0.0, 1.0, 0.0],
                   'PixelSpacing' : [1.0, 1.0],
                   'SliceThickness' : 1.0,
                   'Rows' : 16,
                   'Columns' : 16,
                   'BitsAllocated' : 16,
                   'BitsStored' : 16,
                   'PixelRepresentation' : 0,
                   'SamplesPerPixel' : 1,
                   'PhotometricInterpretation': 'MONOCHROME2',
                  }

def make_dicom(attrs=None, pix_val=1):
    '''Build a mock DICOM dataset for testing purposes'''
    ds = pydicom.dataset.Dataset()
    if attrs is not None:
        for attr_name, attr_val in attrs.items():
            setattr(ds, attr_name, deepcopy(attr_val))
    for attr_name, attr_val in def_dicom_attrs.items():
        if not hasattr(ds, attr_name):
            setattr(ds, attr_name, deepcopy(attr_val))
    if not hasattr(ds, 'PixelData'):
        if ds.PixelRepresentation == 0:
            arr_dtype = np.uint16
        else:
            arr_dtype = np.int16
        arr = np.empty((ds.Rows, ds.Columns), dtype=arr_dtype)
        arr[:, :] = pix_val
        ds.PixelData = arr.tobytes()
    return ds

def test_key_regex_filter():
        filt = dcmstack.make_key_regex_filter(['test', 'another'],
                                              ['2', 'another test'])
        assert(filt('test', 1))
        assert(filt('test another', 1))
        assert(filt('another tes', 1))
        assert(not filt('test2', 1))
        assert(not filt('2 another', 1))
        assert(not filt('another test', 1))

class TestReorderVoxels(object):
    def setup_method(self, method):
        self.vox_array = np.arange(16).reshape((2, 2, 2, 2))
        self.affine = np.eye(4)

    def test_invalid_vox_order(self):
        with pytest.raises(ValueError):
            dcmstack.reorder_voxels(self.vox_array, self.affine, 'lra')
        with pytest.raises(ValueError):
            dcmstack.reorder_voxels(self.vox_array, self.affine, 'rpil')
        with pytest.raises(ValueError):
            dcmstack.reorder_voxels(self.vox_array, self.affine, 'lrz')

    def test_invalid_vox_array(self):
        with pytest.raises(ValueError):
            dcmstack.reorder_voxels(np.eye(2), self.affine, 'rpi')

    def test_invalid_affine(self):
        with pytest.raises(ValueError):
            dcmstack.reorder_voxels(self.vox_array, np.eye(3), 'rpi')

    def test_no_op(self):
        vox_order = ''.join(aff2axcodes(self.affine))
        (vox_array,
         affine,
         aff_trans,
         ornt_trans) = dcmstack.reorder_voxels(self.vox_array,
                                               self.affine,
                                               vox_order)
        assert((vox_array == self.vox_array).all())
        assert((affine == self.affine).all())
        assert((aff_trans == np.eye(4)).all())
        assert(np.allclose(ornt_trans, [[0, 1], [1, 1], [2, 1]]))
        assert np.may_share_memory(affine, self.affine) == False

    def test_reorder(self):
        (vox_array,
         affine,
         aff_trans,
         ornt_trans) = dcmstack.reorder_voxels(self.vox_array,
                                               self.affine,
                                               'PRS')
        assert(np.all(vox_array == np.array([[[[4, 5],
                                            [6, 7]],
                                           [[12, 13],
                                            [14,15]]
                                          ],
                                          [[[0, 1],
                                            [2, 3]],
                                           [[8, 9],
                                            [10, 11]]
                                          ]
                                         ]
                                        )
                  )
           )
        assert(np.allclose(affine,
                        np.array([[0,1,0,0],
                                  [-1,0,0,1],
                                  [0,0,1,0],
                                  [0,0,0,1]])
                       )
           )

    def test_aniso_reorder(self):
        self.vox_array = self.vox_array.reshape(2, 4, 2)
        self.affine = np.eye(4)
        (vox_array,
         affine,
         aff_trans,
         ornt_trans) = dcmstack.reorder_voxels(self.vox_array,
                                               self.affine,
                                               'PLS')
        assert(np.allclose(affine,
                        np.array([[0,-1,0,1],
                                  [-1,0,0,3],
                                  [0,0,1,0],
                                  [0,0,0,1]])
                       )
           )


class TestDicomOrdering(object):
    class _MockNiiWrp:
        def __init__(self, meta_dict):
            self._meta_dict = meta_dict
        
        def get_meta(self, key):
            return self._meta_dict.get(key)
        
    def setup_method(self, method):
        self.meta = {'EchoTime' : 2}
        self.nw = TestDicomOrdering._MockNiiWrp(self.meta)

    def test_missing_key(self):
        ordering = dcmstack.DicomOrdering('blah')
        assert ordering.get_ordinate(self.nw) == None

    def test_non_abs(self):
        ordering = dcmstack.DicomOrdering('EchoTime')
        assert ordering.get_ordinate(self.nw) == self.meta['EchoTime']

    def test_abs(self):
        abs_order = [1,2,3]
        ordering = dcmstack.DicomOrdering('EchoTime', abs_ordering=abs_order)
        assert ordering.get_ordinate(self.nw) == abs_order.index(self.meta['EchoTime'])

    def test_abs_as_str(self):
        abs_order = ['1','2','3']
        ordering = dcmstack.DicomOrdering('EchoTime',
                                          abs_ordering=abs_order,
                                          abs_as_str=True)
        assert ordering.get_ordinate(self.nw) == abs_order.index(str(self.meta['EchoTime']))

    def test_abs_missing(self):
        abs_order = [1,3]
        ordering = dcmstack.DicomOrdering('EchoTime', abs_ordering=abs_order)
        with pytest.raises(ValueError):
            ordering.get_ordinate(self.nw)

def test_image_collision():
    dcm_path = path.join(test_dir,
                         'data',
                         'dcmstack',
                         '2D_16Echo_qT2',
                         'TE_20_SlcPos_-33.707626341697.dcm')
    dcm = pydicom.dcmread(dcm_path)
    stack = dcmstack.DicomStack('EchoTime')
    stack.add_dcm(dcm)
    with pytest.raises(dcmstack.ImageCollisionError):
        stack.add_dcm(dcm)


class TestIncongruentImage(object):
    def setup_method(self, method):
        self.dcm = make_dicom()
        self.stack = dcmstack.DicomStack()
        self.stack.add_dcm(self.dcm)

    def _chk_raises(self, ds):
        with pytest.raises(dcmstack.IncongruentImageError):
            self.stack.add_dcm(ds)

    def test_rows(self):
        ds = make_dicom({"Rows": self.dcm.Rows + 1})
        self._chk_raises(ds)

    def test_columns(self):
        ds = make_dicom({"Columns": self.dcm.Columns + 1})
        self._chk_raises(ds)

    def test_pix_space(self):
        ds = make_dicom(
            {"PixelSpacing": [self.dcm.PixelSpacing[0] * 2, self.dcm.PixelSpacing[1]]}
        )
        self._chk_raises(ds)

    def test_close_pix_space(self):
        ds = make_dicom(
            {"PixelSpacing": [self.dcm.PixelSpacing[0] + 1e-7, self.dcm.PixelSpacing[1]]}
        )
        # Shouldn't raise
        self.stack.add_dcm(ds)

    def test_orientation(self):
        ds = make_dicom({"ImageOrientationPatient": [1., 0., 0., 0., 0., 1.]})
        self._chk_raises(ds)

    def test_close_orientation(self):
        ds = make_dicom({"ImageOrientationPatient": [1. - 1e-8, 0., 0., 0., 1., 0.]})
        # Shouldn't raise
        self.stack.add_dcm(ds)


class TestInvalidStack(object):
    def setup_method(self, method):
        data_dir = path.join(test_dir,
                             'data',
                             'dcmstack',
                             '2D_16Echo_qT2')
        self.inputs = [pydicom.dcmread(path.join(data_dir, fn))
                       for fn in ('TE_20_SlcPos_-33.707626341697.dcm',
                                  'TE_20_SlcPos_-23.207628249046.dcm',
                                  'TE_40_SlcPos_-33.707626341697.dcm',
                                  'TE_60_SlcPos_-23.207628249046.dcm',
                                  'TE_20_SlcPos_-2.2076272953718.dcm'
                                  )
                      ]

    def _chk(self):
        with pytest.raises(dcmstack.InvalidStackError):
            self.stack.get_shape()
        with pytest.raises(dcmstack.InvalidStackError):
                      self.stack.get_affine()
        with pytest.raises(dcmstack.InvalidStackError):
                      self.stack.get_data()
        with pytest.raises(dcmstack.InvalidStackError):
                      self.stack.to_nifti()

    def test_empty(self):
        self.stack = dcmstack.DicomStack()
        self._chk()

    def test_missing_slice(self):
        self.stack = dcmstack.DicomStack()
        self.stack.add_dcm(self.inputs[0])
        self.stack.add_dcm(self.inputs[1])
        self.stack.add_dcm(self.inputs[4])
        self._chk()

    def test_wrong_number_of_files(self):
        self.stack = dcmstack.DicomStack(time_order='EchoTime')
        self.stack.add_dcm(self.inputs[0])
        self.stack.add_dcm(self.inputs[1])
        self.stack.add_dcm(self.inputs[2])
        self._chk()

    def test_vector_var_over_vol(self):
        self.stack = dcmstack.DicomStack(vector_order='EchoTime')
        self.stack.add_dcm(self.inputs[0])
        self.stack.add_dcm(self.inputs[1])
        self.stack.add_dcm(self.inputs[2])
        self.stack.add_dcm(self.inputs[3])
        self._chk()

class TestGetShape(object):
    def setup_method(self, method):
        data_dir = path.join(test_dir,
                             'data',
                             'dcmstack',
                             '2D_16Echo_qT2')
        self.inputs = [pydicom.dcmread(path.join(data_dir, fn))
                       for fn in ('TE_40_SlcPos_-33.707626341697.dcm',
                                  'TE_40_SlcPos_-23.207628249046.dcm',
                                  'TE_60_SlcPos_-33.707626341697.dcm',
                                  'TE_60_SlcPos_-23.207628249046.dcm',
                                  )
                      ]

    def test_single_slice(self):
        stack = dcmstack.DicomStack()
        stack.add_dcm(self.inputs[0])
        shape = stack.shape
        assert shape == (192, 192, 1)

    def test_three_dim(self):
        stack = dcmstack.DicomStack()
        stack.add_dcm(self.inputs[0])
        stack.add_dcm(self.inputs[1])
        shape = stack.shape
        assert shape == (192, 192, 2)

    def test_four_dim(self):
        stack = dcmstack.DicomStack(time_order='EchoTime')
        stack.add_dcm(self.inputs[0])
        stack.add_dcm(self.inputs[1])
        stack.add_dcm(self.inputs[2])
        stack.add_dcm(self.inputs[3])
        shape = stack.shape
        assert shape == (192, 192, 2, 2)

    def test_five_dim(self):
        stack = dcmstack.DicomStack(vector_order='EchoTime')
        stack.add_dcm(self.inputs[0])
        stack.add_dcm(self.inputs[1])
        stack.add_dcm(self.inputs[2])
        stack.add_dcm(self.inputs[3])
        shape = stack.shape
        assert shape == (192, 192, 2, 1, 2)

class TestGuessDim(object):
    def setup_method(self, method):
        data_dir = path.join(test_dir,
                             'data',
                             'dcmstack',
                             '2D_16Echo_qT2')
        self.inputs = [pydicom.dcmread(path.join(data_dir, fn))
                       for fn in ('TE_40_SlcPos_-33.707626341697.dcm',
                                  'TE_40_SlcPos_-23.207628249046.dcm',
                                  'TE_60_SlcPos_-33.707626341697.dcm',
                                  'TE_60_SlcPos_-23.207628249046.dcm',
                                  )
                      ]
        for in_dcm in self.inputs:
            for key in dcmstack.DicomStack.SORT_GUESSES:
                if hasattr(in_dcm, key):
                    delattr(in_dcm, key)

    def _get_vr_ord(self, key, ordinate):
        try:
            tag = keyword_dict[key]
        except KeyError:
            return ordinate
        vr = dictionary_VR(tag)
        if vr == 'TM':
            return '%06d.000000' % ordinate
        else:
            return ordinate

    def test_single_guess(self):
        #Test situations where there is only one possible correct guess
        for key in dcmstack.DicomStack.SORT_GUESSES:
            stack = dcmstack.DicomStack()
            for idx, in_dcm in enumerate(self.inputs):
                setattr(in_dcm, key, self._get_vr_ord(key, idx))
                stack.add_dcm(in_dcm)
            assert stack.shape == (192, 192, 2, 2)
            for in_dcm in self.inputs:
                delattr(in_dcm, key)

    def test_wrong_guess_first(self):
        #Test situations where the initial guesses are wrong
        stack = dcmstack.DicomStack()
        for key in dcmstack.DicomStack.SORT_GUESSES[:-1]:
            for in_dcm in self.inputs:
                setattr(in_dcm, key, self._get_vr_ord(key, 0))
        for idx, in_dcm in enumerate(self.inputs):
            setattr(in_dcm,
                    dcmstack.DicomStack.SORT_GUESSES[-1],
                    self._get_vr_ord(key, idx) )
            stack.add_dcm(in_dcm)
        assert stack.shape == (192, 192, 2, 2)

class TestGetData(object):
    def setup_method(self, method):
        data_dir = path.join(test_dir,
                             'data',
                             'dcmstack',
                             '2D_16Echo_qT2')
        self.inputs = [pydicom.dcmread(path.join(data_dir, fn))
                       for fn in ('TE_40_SlcPos_-33.707626341697.dcm',
                                  'TE_40_SlcPos_-23.207628249046.dcm',
                                  'TE_60_SlcPos_-33.707626341697.dcm',
                                  'TE_60_SlcPos_-23.207628249046.dcm',
                                  )
                      ]

    def test_single_slice(self):
        stack = dcmstack.DicomStack()
        stack.add_dcm(self.inputs[0])
        data = stack.get_data()
        assert data.shape == stack.shape
        assert sha256(data).hexdigest() == \
            '15cfa107ca73810a1c97f1c1872a7a4a05808ba6147e039cef3f63fa08735f5d'

    def test_three_dim(self):
        stack = dcmstack.DicomStack()
        stack.add_dcm(self.inputs[0])
        stack.add_dcm(self.inputs[1])
        data = stack.get_data()
        assert data.shape == stack.shape
        assert sha256(data).hexdigest() == \
            'ab5225fdbedceeea3442b2c9387e1abcbf398c71f525e0017251849c3cfbf49c'

    def test_four_dim(self):
        stack = dcmstack.DicomStack(time_order='EchoTime')
        stack.add_dcm(self.inputs[0])
        stack.add_dcm(self.inputs[1])
        stack.add_dcm(self.inputs[2])
        stack.add_dcm(self.inputs[3])
        data = stack.get_data()
        assert data.shape == stack.shape
        assert sha256(data).hexdigest() == \
            'bb3639a6ece13dc9a11d65f1b09ab3ccaed63b22dcf0f96fb5d3dd8805cc7b8a'

    def test_five_dim(self):
        stack = dcmstack.DicomStack(vector_order='EchoTime')
        stack.add_dcm(self.inputs[0])
        stack.add_dcm(self.inputs[1])
        stack.add_dcm(self.inputs[2])
        stack.add_dcm(self.inputs[3])
        data = stack.get_data()
        assert data.shape == stack.shape
        assert sha256(data).hexdigest() == \
            'bb3639a6ece13dc9a11d65f1b09ab3ccaed63b22dcf0f96fb5d3dd8805cc7b8a'
    
    def test_data_scaling(self):
        inputs = deepcopy(self.inputs)
        for input in inputs:
            input.RescaleSlope = 10.0
        stack = dcmstack.DicomStack(vector_order='EchoTime')
        stack.add_dcm(inputs[0])
        stack.add_dcm(inputs[1])
        stack.add_dcm(inputs[2])
        stack.add_dcm(inputs[3])
        data = stack.get_data()
        unscl_data = stack.get_data(scaled=False)
        assert data.dtype == np.float32
        assert unscl_data.dtype == np.int16
        assert np.allclose(data, unscl_data*10.0)


class TestGetAffine(object):
    def setup_method(self, method):
        self.data_dir = path.join(test_dir,
                             'data',
                             'dcmstack',
                             '2D_16Echo_qT2')
        self.inputs = [pydicom.dcmread(path.join(self.data_dir, fn))
                       for fn in ('TE_20_SlcPos_-33.707626341697.dcm',
                                  'TE_20_SlcPos_-23.207628249046.dcm'
                                 )
                      ]

    def test_single_slice(self):
        stack = dcmstack.DicomStack()
        stack.add_dcm(self.inputs[0])
        affine = stack.affine
        ref = np.load(path.join(self.data_dir, 'single_slice_aff.npy'))
        assert(np.allclose(affine, ref))

    def test_three_dim(self):
        stack = dcmstack.DicomStack()
        stack.add_dcm(self.inputs[0])
        stack.add_dcm(self.inputs[1])
        affine = stack.affine
        ref = np.load(path.join(self.data_dir, 'single_vol_aff.npy'))
        assert(np.allclose(affine, ref))

class TestToNifti(object):

    eq_keys = ['sizeof_hdr',
               'data_type',
               'extents',
               'dim_info',
               'dim',
               'intent_p1',
               'intent_p2',
               'intent_p3',
               'intent_code',
               'datatype',
               'bitpix',
               'slice_start',
               'pixdim',
               'scl_slope',
               'scl_inter',
               'slice_end',
               'slice_code',
               'xyzt_units',
              ]

    close_keys = ['cal_max',
                  'cal_min',
                  'slice_duration',
                  'toffset',
                  'glmax',
                  'glmin',
                  'qform_code',
                  'sform_code',
                  'quatern_b',
                  'quatern_c',
                  'quatern_d',
                  'qoffset_x',
                  'qoffset_y',
                  'qoffset_z',
                  'srow_x',
                  'srow_y',
                  'srow_z',
                 ]

    def setup_method(self, method):
        self.data_dir = path.join(test_dir,
                             'data',
                             'dcmstack',
                             '2D_16Echo_qT2')
        self.inputs = [pydicom.dcmread(path.join(self.data_dir, fn))
                       for fn in ('TE_20_SlcPos_-33.707626341697.dcm',
                                  'TE_20_SlcPos_-23.207628249046.dcm',
                                  'TE_40_SlcPos_-33.707626341697.dcm',
                                  'TE_40_SlcPos_-23.207628249046.dcm',
                                 )
                      ]

    def _build_nii(self, name):
        kwargs = {}
        if name.endswith('_meta'):
            kwargs['embed_meta'] = True
            name = name[:-5]
        if name.endswith('_SAR'):
            kwargs['voxel_order'] = 'SAR'
            name = name[:-4]
        if name == 'two_time_vol':
            stack = dcmstack.DicomStack(time_order='EchoTime')
        elif name == 'two_vector_vol':
            stack = dcmstack.DicomStack(vector_order='EchoTime')
        else:
            stack = dcmstack.DicomStack()
        stack.add_dcm(self.inputs[0])
        if name == 'single_slice':
            return stack.to_nifti(**kwargs)
        stack.add_dcm(self.inputs[1])
        if name == 'single_vol':
            return stack.to_nifti(**kwargs)
        stack.add_dcm(self.inputs[2])
        stack.add_dcm(self.inputs[3])
        if name in ('two_time_vol', 'two_vector_vol'):
            return stack.to_nifti(**kwargs)
        assert False # Unknown name

    def _chk(self, nii, ref_base_fn):
        hdr = nii.header
        ref_nii = nb.load(path.join(self.data_dir, ref_base_fn) + '.nii.gz')
        ref_hdr = ref_nii.header

        for key in self.eq_keys:
            print("Testing key %s" % key)
            v1 = hdr[key]
            v2 = ref_hdr[key]
            try:
                np.testing.assert_equal(v1, v2)
            except AssertionError:
                if key == 'slice_code':
                    warnings.warn(
                        "Random failure due to a slim test volume and absent "
                        "information on slice ordering.  "
                        "See https://github.com/nipy/nibabel/pull/647"
                    )
                    continue
                raise

        for key in self.close_keys:
            print("Testing key %s" % key)
            assert(np.allclose(hdr[key], ref_hdr[key]))

    def test_single_slice(self):
        for tst in ('single_slice', 'single_slice_meta'):
            nii = self._build_nii(tst)
            self._chk(nii, tst)

    def test_single_vol(self):
        for tst in ('single_vol', 'single_vol_meta'):
            nii = self._build_nii(tst)
            self._chk(nii, tst)

    def test_slice_dim_reorient(self):
        for tst in ('single_vol_SAR', 'single_vol_SAR_meta'):
            nii = self._build_nii(tst)
            self._chk(nii, tst)

    def test_two_time_vol(self):
        for tst in ('two_time_vol', 'two_time_vol_meta'):
            nii = self._build_nii(tst)
            self._chk(nii, tst)

    def test_two_vector_vol(self):
        for tst in ('two_vector_vol', 'two_vector_vol_meta'):
            nii = self._build_nii(tst)
            self._chk(nii, tst)


class TestParseAndGroup(object):
    def setup_method(self, method):
        self.data_dir = path.join(test_dir,
                             'data',
                             'dcmstack',
                             '2D_16Echo_qT2')
        self.in_paths = [path.join(self.data_dir, fn)
                         for fn in ('TE_20_SlcPos_-33.707626341697.dcm',
                                    'TE_20_SlcPos_-23.207628249046.dcm',
                                    'TE_40_SlcPos_-33.707626341697.dcm',
                                    'TE_40_SlcPos_-23.207628249046.dcm',
                                   )
                        ]

    def test_default(self):
        res = dcmstack.parse_and_group(self.in_paths)
        assert len(res) == 1
        ds = pydicom.dcmread(self.in_paths[0])
        group_key = list(res.keys())[0]
        for attr_idx, attr in enumerate(dcmstack.DEFAULT_GROUP_KEYS):
            if attr in dcmstack.DEFAULT_CLOSE_KEYS:
                assert(np.allclose(group_key[attr_idx], getattr(ds, attr)))
            else:
                assert group_key[attr_idx] == getattr(ds, attr)


class TestParseAndStack(object):
    def setup_method(self, method):
        self.data_dir = path.join(test_dir,
                             'data',
                             'dcmstack',
                             '2D_16Echo_qT2')
        self.in_paths = [path.join(self.data_dir, fn)
                         for fn in ('TE_20_SlcPos_-33.707626341697.dcm',
                                    'TE_20_SlcPos_-23.207628249046.dcm',
                                    'TE_40_SlcPos_-33.707626341697.dcm',
                                    'TE_40_SlcPos_-23.207628249046.dcm',
                                   )
                        ]

    def test_default(self):
        res = dcmstack.parse_and_stack(self.in_paths)
        assert len(res) == 1
        ds = pydicom.dcmread(self.in_paths[0])
        group_key = list(res.keys())[0]
        for attr_idx, attr in enumerate(dcmstack.DEFAULT_GROUP_KEYS):
            if attr in dcmstack.DEFAULT_CLOSE_KEYS:
                assert(np.allclose(group_key[attr_idx], getattr(ds, attr)))
            else:
                assert group_key[attr_idx] == getattr(ds, attr)
        stack = list(res.values())[0]
        assert(isinstance(stack, dcmstack.DicomStack))
        stack_data = stack.get_data()
        assert stack_data.ndim == 4


def test_fsl_hack():
    ds = make_dicom({'BitsStored': 14, }, 2**14 - 1)
    stack = dcmstack.DicomStack()
    stack.add_dcm(ds)
    data = stack.get_data()    
    assert np.max(data) == (2**14 - 1)
    assert data.dtype == np.int16


def test_pix_overflow():
    ds = make_dicom(pix_val=(2**16 - 1))
    stack = dcmstack.DicomStack()
    stack.add_dcm(ds)
    data = stack.get_data()    
    assert np.max(data) == (2**16 - 1)
    assert data.dtype == np.uint16



""" Information for setup.py that we may also want to access in dcmstack. Can
not import dcmstack.
"""
import sys

_version_major = 0
_version_minor = 10
_version_micro = 0
_version_extra = 'dev0'
__version__ = "%s.%s.%s%s" % (_version_major,
                              _version_minor,
                              _version_micro,
                              _version_extra)

CLASSIFIERS = ["Development Status :: 3 - Alpha",
               "Environment :: Console",
               "Intended Audience :: Science/Research",
               "License :: OSI Approved :: MIT License",
               "Operating System :: OS Independent",
               "Programming Language :: Python",
               "Topic :: Scientific/Engineering"]

description = 'Stack DICOM images into volumes and convert to Nifti'

# Hard dependencies
install_requires = ['pydicom >= 0.9.7',
                    'nibabel >= 5.3.1',
                    'pylibjpeg[all] ; python_version >= "3.8"',
                    'pint',
                   ]

# Extra requirements for building documentation and testing
extras_requires = {'doc':  ["sphinx", "numpydoc"],
                   'test': [
                        'pytest <= 4.6 ; python_version == "2.7"',
                        'pytest ; python_version > "2.7"',
                        'pytest-cov',
                    ],
                  }


NAME                = 'dcmstack'
AUTHOR              = "Brendan Moloney"
AUTHOR_EMAIL        = "moloney@ohsu.edu"
MAINTAINER          = "Brendan Moloney"
MAINTAINER_EMAIL    = "moloney@ohsu.edu"
URL                 = "https://github.com/moloney/dcmstack"
DESCRIPTION         = description
LICENSE             = "MIT license"
CLASSIFIERS         = CLASSIFIERS
PLATFORMS           = "OS Independent"
ISRELEASE           = _version_extra == ''
VERSION             = __version__
INSTALL_REQUIRES    = install_requires
EXTRAS_REQUIRES      = extras_requires
PROVIDES            = ["dcmstack"]



"""Provide some common conversions for DICOM elements"""
import re
from datetime import date, datetime, timedelta, timezone


def tm_to_sec(time_str):
    '''Convert a DICOM time value (value representation of 'TM') to the number
    of seconds past midnight.

    Parameters
    ----------
    time_str : str
        The DICOM time value string

    Returns
    -------
    A floating point representing the number of seconds past midnight
    '''
    #Allow ACR/NEMA style format by removing any colon chars
    time_str = time_str.replace(':', '')
    #Only the hours portion is required
    result = int(time_str[:2]) * 3600
    str_len = len(time_str)
    if str_len > 2:
        result += int(time_str[2:4]) * 60
    if str_len > 4:
        result += float(time_str[4:])
    return float(result)


def da_to_date(date_str):
    '''Convert a DICOM date value (value representation of 'DA') to a python `date`

    Parameters
    ----------
    date_str : str
        The DICOM date value string
    '''
    #Allow ACR/NEMA style format by removing any dot chars
    date_str = date_str.replace('.', '')
    return date(int(date_str[:4]), int(date_str[4:6]), int(date_str[6:]))


def dt_to_datetime(dt_str):
    '''Convert a DICOM datetime value (value representation of 'DT') to a python `datetime`

    Parameters
    ----------
    dt_str : str
        The DICOM datetime value string
    '''
    dt_str, suffix = re.match("([0-9\.]+)([\+\-][0-9]{4})?", dt_str).groups()
    year = int(dt_str[:4])
    month = day = 1
    hour = minute = seconds = microsecs = 0
    if len(dt_str) > 4:
        month = int(dt_str[4:6])
    if len(dt_str) > 6:
        day = int(dt_str[6:8])
    if len(dt_str) > 8:
        hour = int(dt_str[8:10])
    if len(dt_str) > 10:
        minute = int(dt_str[10:12])
    if len(dt_str) > 12:
        seconds = int(dt_str[12:14])
    if len(dt_str) > 14:
        microsecs = int(float(dt_str[14:]) * 1e6)
    tzinfo = None
    if suffix is not None:
        td = timedelta(hours=int(suffix[1:3]), minutes=int(suffix[3:5]))
        if suffix[0] == '-':
            td *= -1
        tzinfo = timezone(td)
    return datetime(year, month, day, hour, minute, seconds, microsecs, tzinfo=tzinfo)



"""
Stack DICOM datasets into volumes. The contents of this module are imported
into the package namespace.
"""
import warnings, re, traceback
from copy import deepcopy

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

from .nifti_wrapper import NiftiWrapper
import numpy as np
import pydicom
from pydicom.datadict import tag_for_keyword
import nibabel as nb
from nibabel.nifti1 import Nifti1Extensions
from nibabel.spatialimages import HeaderDataError
from nibabel.orientations import (io_orientation,
                                  apply_orientation,
                                  inv_ornt_aff,
                                  ornt_transform,
                                  axcodes2ornt)
with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    from nibabel.nicom.dicomwrappers import wrapper_from_data

from .globals import SORT_GUESSES
from . import snd
from .dcmmeta import DcmMetaExtension


def make_key_regex_filter(exclude_res, force_include_res=None):
    '''Make a meta data filter using regular expressions.

    Parameters
    ----------
    exclude_res : sequence
        Sequence of regular expression strings. Any meta data where the key
        matches one of these expressions will be excluded, unless it matches
        one of the `force_include_res`.
    force_include_res : sequence
        Sequence of regular expression strings. Any meta data where the key
        matches one of these expressions will be included.

    Returns
    -------
    A callable which can be passed to `DicomStack` as the `meta_filter`.
    '''
    exclude_re = re.compile('|'.join(['(?:' + regex + ')'
                                      for regex in exclude_res])
                           )
    include_re = None
    if force_include_res:
        include_re = re.compile('|'.join(['(?:' + regex + ')'
                                          for regex in force_include_res])
                               )

    def key_regex_filter(key, value):
        return (exclude_re.search(key) and
                not (include_re and include_re.search(key)))
    return key_regex_filter


default_key_excl_res = ['Patient',
                        'Physician',
                        'Operator',
                        'Date',
                        'Birth',
                        'Address',
                        'Institution',
                        'Station',
                        'SiteName',
                        'Age',
                        'Comment',
                        'Phone',
                        'Telephone',
                        'Insurance',
                        'Religious',
                        'Language',
                        'Military',
                        'MedicalRecord',
                        'Ethnic',
                        'Occupation',
                        'Unknown',
                        'PrivateTagData',
                        'UID',
                        'StudyDescription',
                        'DeviceSerialNumber',
                        'ReferencedImageSequence',
                        'RequestedProcedureDescription',
                        'PerformedProcedureStepDescription',
                        'PerformedProcedureStepID',
                       ]
'''A list of regexes passed to `make_key_regex_filter` as `exclude_res` to
create the `default_meta_filter`.'''


default_key_incl_res = ['ImageOrientationPatient',
                        'ImagePositionPatient',
                       ]
'''A list of regexes passed to `make_key_regex_filter` as `force_include_res`
to create the `default_meta_filter`.'''


default_meta_filter = make_key_regex_filter(default_key_excl_res,
                                            default_key_incl_res)
'''Default meta_filter for `DicomStack`.'''


def reorder_voxels(vox_array, affine, voxel_order):
    '''Reorder the given voxel array and corresponding affine.

    Parameters
    ----------
    vox_array : array
        The array of voxel data

    affine : array
        The affine for mapping voxel indices to Nifti patient space

    voxel_order : str
        A three character code specifing the desired ending point for rows,
        columns, and slices in terms of the orthogonal axes of patient space:
        (l)eft, (r)ight, (a)nterior, (p)osterior, (s)uperior, and (i)nferior.

    Returns
    -------
    out_vox : array
        An updated view of vox_array.

    out_aff : array
        A new array with the updated affine

    reorient_transform : array
        The transform used to update the affine.

    ornt_trans : tuple
        The orientation transform used to update the orientation.

    '''
    #Check if voxel_order is valid
    voxel_order = voxel_order.upper()
    if len(voxel_order) != 3:
        raise ValueError('The voxel_order must contain three characters')
    dcm_axes = ['LR', 'AP', 'SI']
    for char in voxel_order:
        if not char in 'LRAPSI':
            raise ValueError('The characters in voxel_order must be one '
                             'of: L,R,A,P,I,S')
        for idx, axis in enumerate(dcm_axes):
            if char in axis:
                del dcm_axes[idx]
    if len(dcm_axes) != 0:
        raise ValueError('No character in voxel_order corresponding to '
                         'axes: %s' % dcm_axes)

    #Check the vox_array and affine have correct shape/size
    if len(vox_array.shape) < 3:
        raise ValueError('The vox_array must be at least three dimensional')
    if affine.shape != (4, 4):
        raise ValueError('The affine must be 4x4')

    #Pull the current index directions from the affine
    orig_ornt = io_orientation(affine)
    new_ornt = axcodes2ornt(voxel_order)
    ornt_trans = ornt_transform(orig_ornt, new_ornt)
    orig_shape = vox_array.shape
    vox_array = apply_orientation(vox_array, ornt_trans)
    aff_trans = inv_ornt_aff(ornt_trans, orig_shape)
    affine = np.dot(affine, aff_trans)

    return (vox_array, affine, aff_trans, ornt_trans)


class IncongruentImageError(Exception):
    def __init__(self, msg):
        '''An exception denoting that a DICOM with incorrect size or orientation
        was passed to `DicomStack.add_dcm`.'''
        self.msg = msg

    def __str__(self):
        return "The image is not congruent to the existing stack: %s" % self.msg


class ImageCollisionError(Exception):
    '''An exception denoting that a DICOM which collides with one already in
    the stack was passed to a `DicomStack.add_dcm`.'''
    def __str__(self):
        return "The image collides with one already in the stack"


class NonImageDataSetError(Exception):
    '''The DICOM doesn't contain image data'''
    def __str__(self):
        return "The DICOM doesn't contain an image"


class DicomConversionError(Exception):
    def __str__(self):
        return "Unable to convert DICOM"
    

class InvalidStackError(Exception):
    def __init__(self, msg):
        '''An exception denoting that a `DicomStack` is not currently valid'''
        self.msg = msg

    def __str__(self):
        return "The DICOM stack is not valid: %s" % self.msg


class DicomOrdering(object):
    '''Object defining an ordering for a set of dicom datasets. Create a
    DicomOrdering with the given DICOM element keyword.

    Parameters
    ----------
    key : str
        The DICOM keyword to use for ordering the datasets

    abs_ordering : sequence
        A sequence specifying the absolute order for values corresponding
        to the `key`. Instead of ordering by the value associated with the
        `key`, the index of the value in this sequence will be used.

    abs_as_str : bool
        If true, the values will be converted to strings before looking up
        the index in `abs_ordering`.
    '''

    def __init__(self, key, abs_ordering=None, abs_as_str=False):
        self.key = key
        self.abs_ordering = abs_ordering
        self.abs_as_str = abs_as_str

    def get_ordinate(self, nw):
        '''Get the ordinate for the given DICOM data set.

        Parameters
        ----------
        ds : dict like
            The DICOM data set we want the ordinate of. Should allow
            dict like access where DICOM keywords return the corresponing
            value.

        Returns
        -------
        An ordinate for the data set. If `abs_ordering` is None then this will
        just be the value for the keyword `key`. Otherwise it will be an
        integer.
        '''
        val = nw.get_meta(self.key)
        if self.abs_ordering:
            if self.abs_as_str:
                val = str(val)
            return self.abs_ordering.index(val)
        return val


DEFAULT_GROUP_KEYS =  ('SeriesInstanceUID',
                       'SeriesNumber',
                       'ProtocolName',
                       'ImageOrientationPatient')
'''Default keys for grouping DICOM files that belong in the same
multi-dimensional array together.'''


DEFAULT_CLOSE_KEYS = ('ImageOrientationPatient',)
'''Default keys needing np.allclose instead of equaulity testing when grouping 
'''


_pix_tags = tuple(tag_for_keyword(x) for x in ('PixelData', 'FloatPixelData', 'DoubleFloatPixelData'))


def is_image(dcm):
    '''Test if the data set is an image'''
    tags = set(dcm.keys())
    if any(t in tags for t in _pix_tags):
        return True
    return False


class DicomStack(object):
    '''Defines a method for stacking together DICOM data sets into a multi
    dimensional volume.

    Tailored towards creating NiftiImage output, but can also just create numpy
    arrays. Can summarize all of the meta data from the input DICOM data sets
    into a Nifti header extension (see `dcmmeta.DcmMetaExtension`).

    Parameters
    ----------
    time_order : str or DicomOrdering
        The DICOM keyword or DicomOrdering object specifying how to order
        the DICOM data sets along the time dimension.

    vector_order : str or DicomOrdering
        The DICOM keyword or DicomOrdering object specifying how to order
        the DICOM data sets along the vector dimension.

    meta_filter : callable
        A callable that takes a meta data key and value, and returns True if
        that meta data element should be excluded from the DcmMeta extension.

    Notes
    -----
    If both time_order and vector_order are None, the time_order will be
    guessed based off the data sets.
    '''
    OTHER_REQ_ATTRS = (
        'Rows',
        'Columns',
        'PixelSpacing',
        'ImageOrientationPatient',
        'InPlanePhaseEncodingDirection',
        'BitsStored',
        'SharedFunctionalGroupsSequence',
        'PerFrameFunctionalGroupsSequence',
    )

    SORT_GUESSES = SORT_GUESSES[:]

    def __init__(self, time_order=None, vector_order=None, meta_filter=None):
        if isinstance(time_order, str):
            self._time_order = DicomOrdering(time_order)
        else:
            self._time_order = time_order
        if isinstance(vector_order, str):
            self._vector_order = DicomOrdering(vector_order)
        else:
            self._vector_order = vector_order
        if meta_filter is None:
            self._meta_filter = default_meta_filter
        else:
            self._meta_filter = meta_filter
        #Sets all the state variables to their defaults
        self.clear()

    @classmethod
    def get_min_req_meta(cls):
        """Returns `(min_dicom, min_trans)`, the min required DICOM keys and translators 
        """
        min_keys = snd.get_snd_sources()
        for key in DEFAULT_GROUP_KEYS + cls.SORT_GUESSES + cls.OTHER_REQ_ATTRS:
            if key.startswith("SND."):
                continue
            min_keys.add(key)
        min_dicom = []
        min_trans = set()
        for key in min_keys:
            toks = key.split(".")
            if len(toks) == 1:
                min_dicom.append(key)
            else:
                min_trans.add(toks[0])
        return min_dicom, min_trans

    def _chk_congruent(self, in_wrp):
        if not self._ref_input is None:
            if self._ref_input.nii_img.shape != in_wrp.nii_img.shape:
                raise IncongruentImageError("Mismatch in data array shapes")
            if not np.allclose(
                self._ref_input.nii_img.affine[:3, :3], 
                in_wrp.nii_img.affine[:3, :3], 
                atol=5e-5,
            ):
                raise IncongruentImageError("Mismatch in affine transforms")

    def add_dcm(self, dcm, meta=None):
        '''Add a pydicom dataset to the stack.

        Parameters
        ----------
        dcm : pydicom.dataset.Dataset
            The dataset being added to the stack

        meta : dict
            The extracted meta data for the DICOM data set `dcm`. If None just the 
            minimal needed meta data will be extracted

        Raises
        ------
        IncongruentImageError
            The provided `dcm` does not match the orientation or dimensions of
            those already in the stack.

        ImageCollisionError
            The provided `dcm` has the same slice location and time/vector
            values.
        
        NonImageDataSetError
            The provided `dcm` doesn't contain image data
        '''
        if not is_image(dcm):
            raise NonImageDataSetError()
        # Extract at least some minimal meta data
        if meta is None:
            from .extract import EXTRACTORS, ExtractionLevel
            meta = EXTRACTORS[ExtractionLevel.MINIMAL](dcm)
        # Covert to DicomWrapper and NiftiWrapper
        dw = wrapper_from_data(dcm)
        nii_wrp = NiftiWrapper.from_dicom_wrapper(dw, meta)
        # TODO: Look into if / why this check is needed?
        if nii_wrp is None:
            raise DicomConversionError()
        # Inject our slightly normalized DICOM meta data
        snd.inject(nii_wrp.meta_ext)
        # Sanity check against existing inputs, then pull some data we will need later
        self._chk_congruent(nii_wrp)
        slp, inter = nii_wrp.nii_img.header.get_slope_inter()
        slp = 1 if slp is None else slp
        inter = 0 if inter is None else inter
        self._slope_inter.add((slp, inter))
        self._phase_enc_dirs.add(nii_wrp.get_meta('InPlanePhaseEncodingDirection'))
        self._repetition_times.add(nii_wrp.get_meta('RepetitionTime'))
        # Pull the info used for sorting if the keys to use are already specified
        slice_pos = dw.slice_indicator
        self._slice_pos_vals.add(slice_pos)
        time_val = None
        if self._time_order:
            time_val = self._time_order.get_ordinate(nii_wrp)
        self._time_vals.add(time_val)
        vector_val = None
        if self._vector_order:
            vector_val = self._vector_order.get_ordinate(nii_wrp)
        self._vector_vals.add(vector_val)
        sorting_tuple = (vector_val, time_val, slice_pos)
        # Also check for collisions if we already know how to sort
        if ((not self._time_order is None or
             not self._vector_order is None) and
            sorting_tuple in self._sorting_tuples
           ):
            raise ImageCollisionError()
        # Add image to our internal set
        self._sorting_tuples.add(sorting_tuple)
        self._files_info.append((nii_wrp, sorting_tuple))
        if self._ref_input is None:
            self._ref_input = nii_wrp
        #Set the dirty flags
        self._shape_dirty = True
        self._meta_dirty = True

    def clear(self):
        '''Remove any DICOM datasets from the stack.'''
        self._slice_pos_vals = set()
        self._time_vals = set()
        self._vector_vals = set()
        self._sorting_tuples = set()
        self._phase_enc_dirs = set()
        self._repetition_times = set()
        self._slope_inter = set()
        self._ref_input = None
        self._shape_dirty = True
        self._shape = None
        self._meta_dirty = True
        self._meta = None
        self._files_info = []

    def _chk_order(self, slice_positions, files_per_vol, num_volumes,
                   num_time_points, num_vec_comps):
        #Sort the files
        self._files_info.sort(key=lambda x: x[1])
        if files_per_vol > 1:
            for vol_idx in range(num_volumes):
                start_slice = vol_idx * files_per_vol
                end_slice = start_slice + files_per_vol
                self._files_info[start_slice:end_slice] = \
                    sorted(self._files_info[start_slice:end_slice],
                           key=lambda x: x[1][-1])
        #Do a thorough check for correctness
        for vec_idx in range(num_vec_comps):
            file_idx = vec_idx*num_time_points*files_per_vol
            curr_vec_val = self._files_info[file_idx][1][0]
            for time_idx in range(num_time_points):
                for slice_idx in range(files_per_vol):
                    file_idx = (vec_idx*num_time_points*files_per_vol +
                                time_idx*files_per_vol + slice_idx)
                    file_info = self._files_info[file_idx]
                    if file_info[1][0] != curr_vec_val:
                        raise InvalidStackError("Not enough images with the " +
                                                "vector value of " +
                                                str(curr_vec_val))
                    if (file_info[1][2] != slice_positions[slice_idx]):
                        if (file_info[1][2] == slice_positions[slice_idx-1]):
                            error_msg = ["Duplicate slice position"]
                        else:
                            error_msg = ["Missing slice position"]
                        error_msg.append(" at slice index %d" % slice_idx)
                        if num_time_points > 1:
                            error_msg.append(' in time point %d' % time_idx)
                        if num_vec_comps > 1:
                            error_msg.append(' for vector component %s' %
                                             str(curr_vec_val))
                        raise InvalidStackError(''.join(error_msg))

    def get_shape(self):
        '''Get the shape of the stack.

        Returns
        -------
        A tuple of integers giving the size of the dimensions of the stack.

        Raises
        ------
        InvalidStackError
            The stack is incomplete or invalid.
        '''
        #If the dirty flag is not set, return the cached value
        if not self._shape_dirty:
            return self._shape
        self._guessed_order = None
        #We need at least one file in the stack
        n_files = len(self._files_info)
        if n_files == 0:
            raise InvalidStackError("No files in the stack")
        #Figure out number of files and slices per volume
        files_per_vol = len(self._slice_pos_vals)
        slice_positions = sorted(list(self._slice_pos_vals))
        #If more than one file per volume, check that slice spacing is equal
        if files_per_vol > 1:
            spacings = []
            for idx in range(files_per_vol - 1):
                spacings.append(slice_positions[idx+1] - slice_positions[idx])
            spacings = np.array(spacings)
            avg_spacing = np.mean(spacings)
            if not np.allclose(avg_spacing, spacings, rtol=4e-2):
                raise InvalidStackError("Slice spacings are not consistent")
        #Simple check for an incomplete stack
        if n_files % files_per_vol != 0:
            raise InvalidStackError("Number of files is not an even multiple "
                                    "of the number of unique slice positions.")
        num_volumes = n_files // files_per_vol
        #Figure out the number of vector components and time points
        num_vec_comps = len(self._vector_vals)
        if num_vec_comps > num_volumes:
            raise InvalidStackError("Vector variable varies within volumes")
        if num_volumes % num_vec_comps != 0:
            raise InvalidStackError("Number of volumes not an even multiple "
                                    "of the number of vector components.")
        num_time_points = num_volumes // num_vec_comps
        #If both sort keys are None try to guess
        if (num_volumes > 1 and self._time_order is None and
                self._vector_order is None):
            #Get a list of possible sort orders
            possible_orders = []
            for key in self.SORT_GUESSES:
                vals = set([file_info[0].get_meta(key)
                            for file_info in self._files_info]
                          )
                if any(v is None for v in vals):
                    continue
                if len(vals) == num_volumes or len(vals) == n_files:
                    possible_orders.append(key)
            if len(possible_orders) == 0:
                raise InvalidStackError("Unable to guess key for sorting the "
                                        "fourth dimension")
            #Try out each possible sort order
            for time_order in possible_orders:
                #Update sorting tuples
                for idx in range(n_files):
                    nii_wrp, curr_tuple = self._files_info[idx]
                    self._files_info[idx] = (nii_wrp,
                                             (curr_tuple[0],
                                              nii_wrp[time_order],
                                              curr_tuple[2]
                                             )
                                            )
                #Check the order
                try:
                    self._chk_order(slice_positions,
                                    files_per_vol,
                                    num_volumes,
                                    num_time_points,
                                    num_vec_comps)
                except InvalidStackError:
                    pass
                else:
                    self._guessed_order = DicomOrdering(time_order)
                    break
            else:
                raise InvalidStackError("Unable to guess key for sorting the "
                                        "fourth dimension")
        else: 
            # If at least one sort key was specified, just check the order
            self._chk_order(slice_positions,
                            files_per_vol,
                            num_volumes,
                            num_time_points,
                            num_vec_comps)
        # Stack appears to be valid, build the shape tuple
        file_shape = self._files_info[0][0].nii_img.shape
        vol_shape = list(file_shape)
        if files_per_vol > 1:
            vol_shape[2] = files_per_vol
        shape = vol_shape+ [num_time_points, num_vec_comps]
        if shape[4] == 1:
            shape = shape[:-1]
            if shape[3] == 1:
                shape = shape[:-1]
        self._shape = tuple(shape)

        self._shape_dirty = False
        return self._shape

    shape = property(fget=get_shape)

    def get_data(self, scaled=True):
        '''Get an array of the voxel values.

        Parameters
        ----------
        scaled
            Set to False to get unscaled data

        Returns
        -------
        A numpy array filled with values from the DICOM data sets' pixels.

        Raises
        ------
        InvalidStackError
            The stack is incomplete or invalid.
        '''
        if len(self._files_info) == 0:
            raise InvalidStackError("No valid files in the stack")
        # Determine output dtype
        stack_dtype = self._files_info[0][0].nii_img.dataobj.dtype
        bits_stored = self._files_info[0][0].get_meta('BitsStored', default=16)
        out_dtype = stack_dtype
        if scaled:
            if len(self._slope_inter) > 1 or list(self._slope_inter)[0] != (1.0, 0.0):
                out_dtype = np.float32
            else:
                scaled = False
        # This is a hack to keep fslview happy, it does not like unsigned short
        # data. If less than 16 bits are being used for each pixel we default 
        # to signed short.        
        if out_dtype == np.uint16 and bits_stored < 16:
            out_dtype = np.int16
        # Optimization for single file
        if len(self._files_info) == 1:
            vox_array = np.asarray(
                self._files_info[0][0].nii_img.dataobj, dtype=out_dtype
            )
            if scaled:
                slope, inter = list(self._slope_inter)[0]
                vox_array *= slope
                vox_array += inter
            return vox_array
        # Allocate output array
        stack_shape = self.shape
        stack_shape = tuple(list(stack_shape) + ((5 - len(stack_shape)) * [1]))
        vox_array = np.empty(stack_shape, dtype=out_dtype)
        # Fill the array with data
        n_vols = 1
        if len(stack_shape) > 3:
            n_vols *= stack_shape[3]
        if len(stack_shape) > 4:
            n_vols *= stack_shape[4]
        files_per_vol = len(self._files_info) // n_vols
        file_shape = self._files_info[0][0].nii_img.shape
        for vec_idx in range(stack_shape[4]):
            for time_idx in range(stack_shape[3]):
                if files_per_vol == 1 and file_shape[2] != 1:
                    file_idx = vec_idx*(stack_shape[3]) + time_idx
                    vox_array[:, :, :, time_idx, vec_idx] = \
                        np.asarray(self._files_info[file_idx][0].nii_img.dataobj)
                    if scaled:
                        slope, inter = self._files_info[file_idx][0].nii_img.header.get_slope_inter()
                        vox_array[:, :, :, time_idx, vec_idx] *= slope
                        vox_array[:, :, :, time_idx, vec_idx] += inter
                else:
                    for slice_idx in range(files_per_vol):
                        file_idx = (vec_idx*(stack_shape[3]*stack_shape[2]) +
                                    time_idx*(stack_shape[2]) + slice_idx)
                        vox_array[:, :, slice_idx, time_idx, vec_idx] = \
                             np.asarray(self._files_info[file_idx][0].nii_img.dataobj[:, :, 0])
                        if scaled:
                            slope, inter = self._files_info[file_idx][0].nii_img.header.get_slope_inter()
                            if slope is None:
                                slope = 1.0
                            if inter is None:
                                inter = 0.0
                            vox_array[:, :, slice_idx, time_idx, vec_idx] *= slope
                            vox_array[:, :, slice_idx, time_idx, vec_idx] += inter
        # Trim unused time/vector dimensions
        if stack_shape[4] == 1:
            vox_array = vox_array[...,0]
            if stack_shape[3] == 1:
                vox_array = vox_array[...,0]
        return vox_array

    data = property(fget=get_data)

    def get_affine(self):
        '''Get the affine transform for mapping row/column/slice indices
        to Nifti (RAS) patient space.

        Returns
        -------
        A 4x4 numpy array containing the affine transform.

        Raises
        ------
        InvalidStackError
            The stack is incomplete or invalid.
        '''
        #Figure out the number of three (or two) dimensional volumes
        shape = self.shape
        n_vols = 1
        if len(shape) > 3:
            n_vols *= shape[3]
        if len(shape) > 4:
            n_vols *= shape[4]

        #Figure out the number of files in each volume
        files_per_vol = len(self._files_info) // n_vols

        #Pull the DICOM Patient Space affine from the first input
        aff = self._files_info[0][0].nii_img.affine

        #If there is more than one file per volume, we need to fix slice scaling
        if files_per_vol > 1:
            first_offset = aff[:3, 3]
            second_offset = self._files_info[1][0].nii_img.affine[:3, 3]
            scaled_slc_dir = second_offset - first_offset
            aff[:3, 2] = scaled_slc_dir

        return aff

    affine = property(fget=get_affine)

    def to_nifti(self, voxel_order='LAS', embed_meta=False):
        '''Returns a NiftiImage with the data and affine from the stack.

        Parameters
        ----------
        voxel_order : str
            A three character string repsenting the voxel order in patient
            space (see the function `reorder_voxels`). Can be None or an empty
            string to disable reorientation.

        embed_meta : bool
            If true a dcmmeta.DcmMetaExtension will be embedded in the Nifti
            header.

        Returns
        -------
        A nibabel.nifti1.Nifti1Image created with the stack's data and affine.
        '''
        #Get the voxel data and affine
        pre_scaled = len(self._slope_inter) > 1
        data = self.get_data(scaled=pre_scaled)
        affine = self.affine

        #Figure out the number of three (or two) dimensional volumes
        n_vols = 1
        if len(data.shape) > 3:
            n_vols *= data.shape[3]
        if len(data.shape) > 4:
            n_vols *= data.shape[4]

        vols_per_file = n_vols // len(self._files_info)
        files_per_vol = len(self._files_info) // n_vols

        #Reorder the voxel data if requested
        permutation = [0, 1, 2]
        slice_dim = 2
        reorient_transform = np.eye(4)
        if voxel_order:
            (data,
             affine,
             reorient_transform,
             ornt_trans) = reorder_voxels(data, affine, voxel_order)
            permutation, flips = zip(*ornt_trans)
            permutation = [int(val) for val in permutation]

            #Reverse file order in each volume's files if we flipped slice order
            #This will keep the slice times and meta data order correct
            if files_per_vol > 1 and flips[slice_dim] == -1:
                self._shape_dirty = True
                for vol_idx in range(n_vols):
                    start = vol_idx * files_per_vol
                    stop = start + files_per_vol
                    self._files_info[start:stop] = [self._files_info[idx]
                                                    for idx in range(stop - 1,
                                                                      start - 1,
                                                                      -1)
                                                   ]

            #Update the slice dim
            slice_dim = permutation[2]

        #Create the nifti image using the data array
        nifti_image = nb.Nifti1Image(data, affine)
        nifti_header = nifti_image.header
        if not pre_scaled:
            slope, inter = list(self._slope_inter)[0]
            if (slope, inter) != (1.0, 0.0):
                nifti_header.set_slope_inter(slope, inter)

        #Set the units and dimension info
        nifti_header.set_xyzt_units('mm', 'msec')
        if len(self._repetition_times) == 1 and not None in self._repetition_times:
            nifti_header['pixdim'][4] = list(self._repetition_times)[0]
        dim_info = {'freq' : None, 'phase' : None, 'slice' : slice_dim}
        if len(self._phase_enc_dirs) == 1 and not None in self._phase_enc_dirs:
            phase_dir = list(self._phase_enc_dirs)[0]
            if phase_dir == 'ROW':
                dim_info['phase'] = permutation[1]
                dim_info['freq'] = permutation[0]
            else:
                dim_info['phase'] = permutation[0]
                dim_info['freq'] = permutation[1]
        nifti_header.set_dim_info(**dim_info)
        
        n_slices = data.shape[slice_dim]

        #Set the slice timing header info
        has_acq_time = (self._files_info[0][0].get_meta('SND.AcquisitionTimeStamp') !=
                        None)
        if files_per_vol > 1 and has_acq_time:
            #Pull out the relative slice times for the first volume
            slice_times = np.array([file_info[0]['SND.AcquisitionTimeStamp']
                                    for file_info in self._files_info[:n_slices]]
                                  )
            slice_times -= np.min(slice_times)

            #If there is more than one volume, check if times are consistent
            is_consistent = True
            for vol_idx in range(1, n_vols):
                start_slice = vol_idx * n_slices
                end_slice = start_slice + n_slices
                slices_info = self._files_info[start_slice:end_slice]
                vol_slc_times = \
                    np.array([file_info[0]['SND.AcquisitionTimeStamp']
                              for file_info in slices_info]
                            )
                vol_slc_times -= np.min(vol_slc_times)
                if not np.allclose(slice_times, vol_slc_times):
                    is_consistent = False
                    break

            #If the times are consistent and not all zero, try setting the slice
            #times (sets the slice duration and code if possible).
            if is_consistent and not np.allclose(slice_times, 0.0):
                try:
                    nifti_header.set_slice_times(slice_times)
                except HeaderDataError:
                    pass

        #Embed the meta data extension if requested
        if embed_meta:
            if len(self._files_info) > 1:
                #Build meta data for each volume if needed
                vol_meta = []
                if files_per_vol > 1:
                    for vol_idx in range(n_vols):
                        start_slice = vol_idx * n_slices
                        end_slice = start_slice + n_slices
                        exts = [file_info[0].meta_ext
                                for file_info in self._files_info[start_slice:end_slice]]
                        meta = DcmMetaExtension.from_sequence(exts, 2)
                        vol_meta.append(meta)
                else:
                    vol_meta = [file_info[0].meta_ext
                                for file_info in self._files_info]
                #Build meta data for each time point / vector component
                if len(data.shape) == 5:
                    if data.shape[3] != 1:
                        vec_meta = []
                        for vec_idx in range(data.shape[4]):
                            start_idx = vec_idx * data.shape[3]
                            end_idx = start_idx + data.shape[3]
                            meta = DcmMetaExtension.from_sequence(
                                vol_meta[start_idx:end_idx], 3
                            )
                            vec_meta.append(meta)
                    else:
                        vec_meta = vol_meta
                    meta_ext = DcmMetaExtension.from_sequence(vec_meta, 4)
                elif len(data.shape) == 4:
                    meta_ext = DcmMetaExtension.from_sequence(vol_meta, 3)
                else:
                    meta_ext = vol_meta[0]
                    # TODO(BUG): file_info is not available here!
                    if meta_ext is self._files_info[0][0].meta_ext:
                        meta_ext = deepcopy(meta_ext)
            else:
                meta_ext = self._files_info[0][0].meta_ext
               
            meta_ext.shape = data.shape
            meta_ext.slice_dim = slice_dim
            meta_ext.affine = nifti_header.get_best_affine()
            meta_ext.reorient_transform = reorient_transform
            # Track which attributes were used to sort extra-spatial dimensions
            extra_shape = meta_ext.shape[3:]
            if extra_shape:
                if extra_shape[0] != 1 and meta_ext.time_dim is None:
                    ord = self._time_order
                    if ord is None:
                        ord = self._guessed_order
                    if ord is not None:
                        meta_ext.time_dim = ord.key
                if len(extra_shape) == 2 and extra_shape[1] != 1:
                    ord = self._vector_order
                    if ord is not None:
                        meta_ext.vector_dim = ord.key
            #Filter and embed the meta data
            meta_ext.filter_meta(self._meta_filter)
            nifti_header.extensions = Nifti1Extensions([meta_ext])

        nifti_image.update_header()
        return nifti_image

    def to_nifti_wrapper(self, voxel_order=''):
        '''Convienance method. Calls `to_nifti` and returns a `NiftiWrapper`
        generated from the result.
        '''
        return NiftiWrapper(self.to_nifti(voxel_order, True))


def parse_and_group(src_paths, group_by=DEFAULT_GROUP_KEYS, extractor=None,
                    force=False, warn_on_except=False,
                    close_tests=DEFAULT_CLOSE_KEYS):
    '''Parse the given dicom files and group them together. Each group is
    stored as a (list) value in a dict where the key is a tuple of values
    corresponding to the keys in 'group_by'

    Parameters
    ----------
    src_paths : sequence
        A list of paths to the source DICOM files.

    group_by : tuple
        Meta data keys to group data sets with. Any data set with the same
        values for these keys will be grouped together. This tuple of values
        will also be the key in the result dictionary.

    extractor : callable
        Should take a pydicom.dataset.Dataset and return a dictionary of the
        extracted meta data.

    force : bool
        Force reading source files even if they do not appear to be DICOM.

    warn_on_except : bool
        Convert exceptions into warnings, possibly allowing some results to be
        returned.

    close_tests : sequence
        Any `group_by` key listed here is tested with `numpy.allclose` instead
        of straight equality when determining group membership.

    Returns
    -------
    groups : dict
        A dict mapping tuples of values (corresponding to 'group_by') to groups
        of data sets. Each element in the list is a tuple containing the dicom
        object, the parsed meta data, and the filename.
    '''
    if extractor is None:
        from .extract import EXTRACTORS, ExtractionLevel
        extractor = EXTRACTORS[ExtractionLevel.MINIMAL]

    results = {}
    close_elems = {}
    for dcm_path in src_paths:
        #Read the DICOM file
        try:
            dcm = pydicom.dcmread(dcm_path, force=force, defer_size=32)
        except Exception as e:
            if warn_on_except:
                warnings.warn('Error reading file %s: %s' % (dcm_path, str(e)))
                continue
            else:
                raise

        # Warn and skip non-image data sets
        if not is_image(dcm):
            warnings.warn("Skipping non-image data set: %s" % dcm_path)
            continue

        #Extract the meta data and group
        meta = extractor(dcm)
        key_list = [] # Values from group_by elems with equality testing
        close_list = [] # Values from group_by elems with np.allclose testing
        for grp_key in group_by:
            key_elem = meta.get(grp_key)
            if isinstance(key_elem, list) or isinstance(key_elem, pydicom.multival.MultiValue):
                key_elem = tuple(key_elem)
            if grp_key in close_tests:
                close_list.append(key_elem)
            else:
                key_list.append(key_elem)

        # Initially each key has multiple sub_results (corresponding to
        # different values of the "close" keys)
        key = tuple(key_list)
        if not key in results:
            results[key] = [(close_list, [(dcm, meta, dcm_path)])]
        else:
            # Look for a matching sub_result
            for c_list, sub_res in results[key]:
                for c_idx, c_val in enumerate(c_list):
                    if not (
                        (c_val is None and close_list[c_idx] is None) or
                        np.allclose(c_val, close_list[c_idx], atol=5e-5)
                    ):
                        break
                else:
                    sub_res.append((dcm, meta, dcm_path))
                    break
            else:
                # No match found, append another sub result
                results[key].append((close_list, [(dcm, meta, dcm_path)]))

    # Unpack sub results, using the canonical value for the close keys
    full_results = {}
    for eq_key, sub_res_list in results.items():
        for close_key, sub_res in sub_res_list:
            full_key = []
            eq_idx = 0
            close_idx = 0
            for grp_key in group_by:
                if grp_key in close_tests:
                    full_key.append(close_key[close_idx])
                    close_idx += 1
                else:
                    full_key.append(eq_key[eq_idx])
                    eq_idx += 1
            full_key = tuple(full_key)
            full_results[full_key] = sub_res

    return OrderedDict(sorted(full_results.items()))


def stack_group(group, warn_on_except=False, **stack_args):
    result = DicomStack(**stack_args)
    for dcm, meta, fn in group:
        try:
            result.add_dcm(dcm, meta)
        except Exception as e:
            if warn_on_except:
                tb_str = "\n".join(traceback.format_tb(e.__traceback__))
                warnings.warn('Error adding file %s to stack: %s\n%s' %
                              (fn, str(e), tb_str))
            else:
                raise
    return result


def parse_and_stack(src_paths, group_by=DEFAULT_GROUP_KEYS, extractor=None,
                    force=False, warn_on_except=False, **stack_args):
    '''Parse the given dicom files into a dictionary containing one or more
    DicomStack objects.

    Parameters
    ----------
    src_paths : sequence
        A list of paths to the source DICOM files.

    group_by : tuple
        Meta data keys to group data sets with. Any data set with the same
        values for these keys will be grouped together. This tuple of values
        will also be the key in the result dictionary.

    extractor : callable
        Should take a pydicom.dataset.Dataset and return a dictionary of the
        extracted meta data.

    force : bool
        Force reading source files even if they do not appear to be DICOM.

    warn_on_except : bool
        Convert exceptions into warnings, possibly allowing some results to be
        returned.

    stack_args : kwargs
        Keyword arguments to pass to the DicomStack constructor.
    '''
    results = parse_and_group(src_paths,
                              group_by,
                              extractor,
                              force,
                              warn_on_except)

    for key, group in results.items():
        results[key] = stack_group(group, warn_on_except, **stack_args)

    return results



"""
Tests for dcmstack.dcmmeta
"""
import sys
from os import path
from glob import glob

import pytest
import numpy as np
import nibabel as nb
try:
    import pydicom
except ImportError:
    import dicom as pydicom

from . import test_dir, src_dir

from dcmstack import dcmmeta
from dcmstack import nifti_wrapper


def test_is_constant():
    assert(dcmmeta.is_constant([0]))
    assert(dcmmeta.is_constant([0, 0]))
    assert(dcmmeta.is_constant([0, 0, 1, 1], period=2))
    assert(dcmmeta.is_constant([0, 0, 1, 1, 2, 2], period=2))
    assert dcmmeta.is_constant([0, 1]) == False
    assert dcmmeta.is_constant([0, 0, 1, 2], 2) == False
    with pytest.raises(ValueError):
        dcmmeta.is_constant([0, 0, 0], -1)
        dcmmeta.is_constant([0, 0, 0], 1)
        dcmmeta.is_constant([0, 0, 0], 2)
        dcmmeta.is_constant([0, 0, 0], 4)

def test_is_repeating():
    assert(dcmmeta.is_repeating([0, 1, 0, 1], 2))
    assert(dcmmeta.is_repeating([0, 1, 0, 1, 0, 1], 2))
    assert dcmmeta.is_repeating([0, 1, 1, 2], 2) == False
    with pytest.raises(ValueError):
        dcmmeta.is_repeating([0, 1, 0, 1], -1)
        dcmmeta.is_repeating([0, 1, 0, 1], 1)
        dcmmeta.is_repeating([0, 1, 0, 1], 3)
        dcmmeta.is_repeating([0, 1, 0, 1], 4)
        dcmmeta.is_repeating([0, 1, 0, 1], 5)

def test_get_valid_classes():
    ext = dcmmeta.DcmMetaExtension.make_empty((2, 2, 2), np.eye(4))
    assert ext.get_valid_classes() == (('global', 'const'), ('global', 'slices'))

    ext.shape = (2, 2, 2, 2)
    assert ext.get_valid_classes() == (('global', 'const'), ('global', 'slices'), ('time', 'samples'), ('time', 'slices'))

    ext.shape = (2, 2, 2, 1, 2)
    assert ext.get_valid_classes() == (('global', 'const'), ('global', 'slices'), ('vector', 'samples'), ('vector', 'slices'))

    ext.shape = (2, 2, 2, 2, 2)
    assert ext.get_valid_classes() == (('global', 'const'), \
                                      ('global', 'slices'), \
                                      ('time', 'samples'), \
                                      ('time', 'slices'), \
                                      ('vector', 'samples'), \
                                      ('vector', 'slices'))

def test_get_mulitplicity_4d():
    ext = dcmmeta.DcmMetaExtension.make_empty((64, 64, 7, 11),
                                              np.eye(4),
                                              np.eye(4),
                                              2)
    assert ext.get_multiplicity(('global', 'const')) == 1
    assert ext.get_multiplicity(('global', 'slices')) == 7 * 11
    assert ext.get_multiplicity(('time', 'samples')) == 11
    assert ext.get_multiplicity(('time', 'slices')) == 7

def test_get_mulitplicity_4d_vec():
    ext = dcmmeta.DcmMetaExtension.make_empty((64, 64, 7, 1, 11),
                                              np.eye(4),
                                              np.eye(4),
                                              2)
    assert ext.get_multiplicity(('global', 'const')) == 1
    assert ext.get_multiplicity(('global', 'slices')) == 7 * 11
    assert ext.get_multiplicity(('vector', 'samples')) == 11
    assert ext.get_multiplicity(('vector', 'slices')) == 7

def test_get_mulitplicity_5d():
    ext = dcmmeta.DcmMetaExtension.make_empty((64, 64, 7, 11, 13),
                                              np.eye(4),
                                              np.eye(4),
                                              2
                                             )
    assert ext.get_multiplicity(('global', 'const')) == 1
    assert ext.get_multiplicity(('global', 'slices')) == 7 * 11 * 13
    assert ext.get_multiplicity(('time', 'samples')) == 11 * 13
    assert ext.get_multiplicity(('time', 'slices')) == 7
    assert ext.get_multiplicity(('vector', 'samples')) == 13
    assert ext.get_multiplicity(('vector', 'slices')) == 7 * 11

class TestCheckValid(object):
    def setUp(self):
        self.ext = dcmmeta.DcmMetaExtension.make_empty((64, 64, 2, 3, 4),
                                                       np.eye(4),
                                                       np.eye(4),
                                                       2)

    def setup_method(self, setUp):
        self.setUp()

    def test_empty(self):
        self.ext.check_valid()

    def test_const(self):
        self.ext.get_class_dict(('global', 'const'))['ConstTest'] = 2
        self.ext.check_valid()
        del self.ext._content['global']['const']
        with pytest.raises(dcmmeta.InvalidExtensionError):
            self.ext.check_valid()

    def test_global_slices(self):
        cls_dict = self.ext.get_class_dict(('global', 'slices'))
        cls_dict['SliceTest'] = [0] * 23
        with pytest.raises(dcmmeta.InvalidExtensionError):
            self.ext.check_valid()
        cls_dict['SliceTest'] = [0] * 24
        self.ext.check_valid()
        del self.ext._content['global']['slices']
        with pytest.raises(dcmmeta.InvalidExtensionError):
            self.ext.check_valid()

    def test_time_samples(self):
        cls_dict = self.ext.get_class_dict(('time', 'samples'))
        cls_dict['TimeSampleTest'] = [0] * 2
        with pytest.raises(dcmmeta.InvalidExtensionError):
            self.ext.check_valid()
        cls_dict['TimeSampleTest'] = [0] * 3
        with pytest.raises(dcmmeta.InvalidExtensionError):
            self.ext.check_valid()
        cls_dict['TimeSampleTest'] = [0] * 12
        self.ext.check_valid()
        del self.ext._content['time']['samples']
        with pytest.raises(dcmmeta.InvalidExtensionError):
            self.ext.check_valid()

    def test_time_slices(self):
        cls_dict = self.ext.get_class_dict(('time', 'slices'))
        cls_dict['TimeSliceTest'] = [0]
        with pytest.raises(dcmmeta.InvalidExtensionError):
            self.ext.check_valid()
        cls_dict['TimeSliceTest'] = [0] * 2
        self.ext.check_valid()
        del self.ext._content['time']['slices']
        with pytest.raises(dcmmeta.InvalidExtensionError): 
            self.ext.check_valid()

    def test_vector_samples(self):
        cls_dict = self.ext.get_class_dict(('vector', 'samples'))
        cls_dict['VectorSampleTest'] = [0] * 3
        with pytest.raises(dcmmeta.InvalidExtensionError):
            self.ext.check_valid()
        cls_dict['VectorSampleTest'] = [0] * 4
        self.ext.check_valid()
        del self.ext._content['vector']['samples']
        with pytest.raises(dcmmeta.InvalidExtensionError):
            self.ext.check_valid()

    def test_vector_slices(self):
        cls_dict = self.ext.get_class_dict(('vector', 'slices'))
        cls_dict['VectorSliceTest'] = [0] * 5
        with pytest.raises(dcmmeta.InvalidExtensionError):
            self.ext.check_valid()
        cls_dict['VectorSliceTest'] = [0] * 6
        self.ext.check_valid()
        del self.ext._content['vector']['slices']
        with pytest.raises(dcmmeta.InvalidExtensionError):
            self.ext.check_valid()

    def test_invalid_affine(self):
        self.ext._content['dcmmeta_affine'] = np.eye(3).tolist()
        with pytest.raises(dcmmeta.InvalidExtensionError):
            self.ext.check_valid()

    def test_invalid_slice_dim(self):
        self.ext._content['dcmmeta_slice_dim'] = 3
        with pytest.raises(dcmmeta.InvalidExtensionError):
            self.ext.check_valid()
        self.ext._content['dcmmeta_slice_dim'] = -1
        with pytest.raises(dcmmeta.InvalidExtensionError):
            self.ext.check_valid()

    def test_invalid_shape(self):
        self.ext._content['dcmmeta_shape'] = [2, 2]
        with pytest.raises(dcmmeta.InvalidExtensionError):
            self.ext.check_valid()
        self.ext._content['dcmmeta_shape'] = [2, 2, 1, 1, 1, 2]
        with pytest.raises(dcmmeta.InvalidExtensionError):
            self.ext.check_valid()

    def test_multiple_classes(self):
        self.ext.get_class_dict(('global', 'const'))['Test'] = 0
        self.ext.get_class_dict(('time', 'samples'))['Test'] = [0] * 3
        with pytest.raises(dcmmeta.InvalidExtensionError):
            self.ext.check_valid()

def test_dcmmeta_affine():
    ext = dcmmeta.DcmMetaExtension.make_empty((64, 64, 2),
                                              np.diag([1, 2, 3, 4]),
                                              np.eye(4),
                                              2
                                             )
    assert(np.allclose(ext.affine, np.diag([1, 2, 3, 4])))
    with pytest.raises(ValueError):
        setattr(ext, 'affine', np.eye(3))
    ext.affine = np.eye(4)
    assert(np.allclose(ext.affine, np.eye(4)))

def test_dcmmeta_slice_dim():
    ext = dcmmeta.DcmMetaExtension.make_empty((64, 64, 2), np.eye(4))
    assert ext.slice_dim == None
    with pytest.raises(ValueError):
        setattr(ext, 'slice_dim', 3)
    ext.slice_dim = 2
    assert ext.slice_dim == 2

def test_dcmmeta_shape():
    ext = dcmmeta.DcmMetaExtension.make_empty((64, 64, 2), np.eye(4))
    assert ext.shape == (64, 64, 2)
    with pytest.raises(ValueError):
        setattr(ext, 'shape', (64, 64))
    ext.shape = (128, 128, 64)
    assert ext.shape == (128, 128, 64)

def test_dcmmeta_version():
    ext = dcmmeta.DcmMetaExtension.make_empty((64, 64, 2), np.eye(4))
    assert ext.version == dcmmeta._meta_version
    ext.version = 1.0
    assert ext.version == 1.0

def test_dcmmeta_slice_norm():
    ext = dcmmeta.DcmMetaExtension.make_empty((64, 64, 2),
                                              np.eye(4),
                                              np.eye(4),
                                              2)
    assert(np.allclose(ext.slice_normal, [0, 0, 1]))
    ext.slice_dim = 1
    assert(np.allclose(ext.slice_normal, [0, 1, 0]))

def test_dcmmeta_n_slices():
    ext = dcmmeta.DcmMetaExtension.make_empty((64, 64, 2),
                                              np.eye(4),
                                              np.eye(4),
                                              2)
    assert ext.n_slices == 2
    ext.slice_dim = 1
    assert ext.n_slices == 64

class TestGetKeysClassesValues(object):
    def setUp(self):
        self.ext = dcmmeta.DcmMetaExtension.make_empty((64, 64, 2, 3, 4),
                                                       np.eye(4),
                                                       np.eye(4),
                                                       2
                                                      )
        self.keys = []
        for classes in self.ext.get_valid_classes():
            key = '%s_%s_test' % classes
            self.keys.append(key)
            self.ext.get_class_dict(classes)[key] = \
                ([0] * self.ext.get_multiplicity(classes))

    def setup_method(self, setUp):
        self.setUp()

    def test_get_keys(self):
        assert set(self.keys) == set(self.ext.get_keys())

    def test_get_classification(self):
        assert self.ext.get_classification('foo') == None
        for classes in self.ext.get_valid_classes():
            key = '%s_%s_test' % classes
            assert self.ext.get_classification(key) == classes

    def test_get_class_dict(self):
        for classes in self.ext.get_valid_classes():
            key = '%s_%s_test' % classes
            assert(key in self.ext.get_class_dict(classes))

    def test_get_values(self):
        assert self.ext.get_values('foo') == None
        for classes in self.ext.get_valid_classes():
            key = '%s_%s_test' % classes
            assert self.ext.get_values(key) == [0] * self.ext.get_multiplicity(classes)
               

    def test_get_vals_and_class(self):
        assert self.ext.get_values_and_class('foo') == (None, None)
        for classes in self.ext.get_valid_classes():
            key = '%s_%s_test' % classes
            assert(self.ext.get_values_and_class(key) == ([0] * self.ext.get_multiplicity(classes), classes))

class TestFiltering(object):
    def setUp(self):
        self.ext = dcmmeta.DcmMetaExtension.make_empty((64, 64, 2, 3, 4),
                                                       np.eye(4),
                                                       np.eye(4),
                                                       2
                                                      )

        for classes in self.ext.get_valid_classes():
            prefix = '%s_%s_test_' % classes
            self.ext.get_class_dict(classes)[prefix + 'foo'] = \
                    ([0] * self.ext.get_multiplicity(classes))
            self.ext.get_class_dict(classes)[prefix + 'foobaz'] = \
                    ([0] * self.ext.get_multiplicity(classes))

    def setup_method(self, setUp):
        self.setUp()

    def test_filter_all(self):
        self.ext.filter_meta(lambda key, val: 'foo' in key)
        assert len(self.ext.get_keys()) == 0

    def test_filter_some(self):
        self.ext.filter_meta(lambda key, val: key.endswith('baz'))
        keys = self.ext.get_keys()
        for classes in self.ext.get_valid_classes():
            prefix = '%s_%s_test_' % classes
            assert(prefix + 'foo' in keys)
            assert(not prefix + 'foobaz' in keys)

    def test_clear_slices(self):
        self.ext.clear_slice_meta()
        for base_cls, sub_cls in self.ext.get_valid_classes():
            if sub_cls == 'slices':
                assert len(self.ext.get_class_dict((base_cls, sub_cls))) == 0

class TestSimplify(object):
    def setUp(self):
        self.ext = dcmmeta.DcmMetaExtension.make_empty((64, 64, 3, 5, 7),
                                                       np.eye(4),
                                                       np.eye(4),
                                                       2
                                                      )
    def setup_method(self, setUp):
        self.setUp()

    def test_simplify_global_slices(self):
        glob_slc = self.ext.get_class_dict(('global', 'slices'))
        glob_slc['Test1'] = [0] * (3 * 5 * 7)
        glob_slc['Test2'] = []
        for idx in range(7):
            glob_slc['Test2'] += [idx] * (3 * 5)
        glob_slc['Test3'] = []
        for idx in range(5 * 7):
            glob_slc['Test3'] += [idx] * (3)
        glob_slc['Test4'] = []
        for idx in range(7):
            glob_slc['Test4'] += [idx2 for idx2 in range(3*5)]
        glob_slc['Test5'] = []
        for idx in range(7 * 5):
            glob_slc['Test5'] += [idx2 for idx2 in range(3)]
        self.ext.check_valid()

        assert self.ext._simplify('Test1') == True
        assert self.ext.get_classification('Test1') == ('global', 'const')

        assert self.ext._simplify('Test2') == True
        assert self.ext.get_classification('Test2'), ('vector', 'samples')

        assert self.ext._simplify('Test3') == True
        assert self.ext.get_classification('Test3') == ('time', 'samples')

        assert self.ext._simplify('Test4') == True
        assert self.ext.get_classification('Test4'), ('vector', 'slices')

        assert self.ext._simplify('Test5') == True
        assert self.ext.get_classification('Test5'), ('time', 'slices')

    def test_simplify_vector_slices(self):
        vec_slc = self.ext.get_class_dict(('vector', 'slices'))
        vec_slc['Test1'] = [0] * (3 * 5)
        vec_slc['Test2'] = []
        for time_idx in range(5):
            vec_slc['Test2'] += [time_idx] * 3
        vec_slc['Test3'] = []
        for time_idx in range(5):
            for slc_idx in range(3):
                vec_slc['Test3'] += [slc_idx]
        self.ext.check_valid()

        assert self.ext._simplify('Test1') ==  True
        assert self.ext.get_classification('Test1') == ('global', 'const')

        assert self.ext._simplify('Test2') ==  True
        assert self.ext.get_classification('Test2') == ('time', 'samples')

        assert self.ext._simplify('Test3') ==  True
        assert self.ext.get_classification('Test3') == ('time', 'slices')

    def test_simplify_time_slices(self):
        time_slc = self.ext.get_class_dict(('time', 'slices'))
        time_slc['Test1'] = [0] * 3
        self.ext.check_valid()

        assert self.ext._simplify('Test1') == True
        assert self.ext.get_classification('Test1') == ('global', 'const')

    def test_simplify_time_samples(self):
        time_smp = self.ext.get_class_dict(('time', 'samples'))
        time_smp['Test1'] = [0] * (5 * 7)
        time_smp['Test2'] = []
        for vec_idx in range(7):
            time_smp['Test2'] += [vec_idx] * 5
        self.ext.check_valid()

        assert self.ext._simplify('Test1') == True
        assert self.ext.get_values_and_class('Test1') == (0, ('global', 'const'))

        assert self.ext._simplify('Test2') == True
        assert self.ext.get_classification('Test2') == ('vector', 'samples')

    def test_simplify_vector_samples(self):
        vector_smp = self.ext.get_class_dict(('vector', 'samples'))
        vector_smp['Test1'] = [0] * 7
        self.ext.check_valid()

        assert self.ext._simplify('Test1') == True
        assert self.ext.get_classification('Test1') == ('global', 'const')

def test_simp_sngl_slc_5d():
    ext = dcmmeta.DcmMetaExtension.make_empty((64, 64, 1, 3, 5),
                                              np.eye(4),
                                              np.eye(4),
                                              2
                                             )
    glob_slc = ext.get_class_dict(('global', 'slices'))
    glob_slc['test1'] = list(range(15))
    ext._simplify('test1')
    assert ext.get_values_and_class('test1') == (list(range(15)), ('time','samples'))

class TestGetSubset(object):
    def setUp(self):
        self.ext = dcmmeta.DcmMetaExtension.make_empty((64, 64, 3, 5, 7),
                                                       np.eye(4),
                                                       np.eye(4),
                                                       2
                                                      )

        #Add an element to every classification
        for classes in self.ext.get_valid_classes():
            key = '%s_%s_test' % classes
            mult = self.ext.get_multiplicity(classes)
            self.ext.get_class_dict(classes)[key] = list(range(mult))

    def setup_method(self, setUp):
        self.setUp()

    def test_slice_subset(self):
        for slc_idx in range(self.ext.n_slices):
            sub = self.ext.get_subset(2, slc_idx)
            sub.check_valid()

            for classes in self.ext.get_valid_classes():
                key = '%s_%s_test' % classes
                if classes == ('time', 'slices'):
                    assert sub.get_values_and_class(key) ==(slc_idx, ('global', 'const'))

                elif classes == ('vector', 'slices'):
                    assert sub.get_values_and_class(key) == ((list(range(slc_idx, (3 * 5), 3)) * 7), ('time', 'samples'))

                elif classes == ('global', 'slices'):
                    assert sub.get_values_and_class(key) == (list(range(slc_idx, (3 * 5 * 7), 3)), ('time', 'samples'))
                else:
                    assert sub.get_values_and_class(key) == self.ext.get_values_and_class(key)

    def test_slice_subset_simplify(self):
        vals = []
        for time_idx in range(5):
            for slice_idx in range(3):
                if slice_idx == 1:
                    vals.append(1)
                else:
                    vals.append(time_idx)
        self.ext.get_class_dict(('vector', 'slices'))['vec_slc_to_const'] = vals

        vals = []
        for vector_idx in range(7):
            for time_idx in range(5):
                for slice_idx in range(3):
                    if slice_idx == 1:
                        vals.append(1)
                    else:
                        vals.append(vector_idx)
        self.ext.get_class_dict(('global', 'slices'))['glb_slc_to_const'] = \
            vals
        self.ext.check_valid()

        sub = self.ext.get_subset(2, 1)
        sub.check_valid()
        assert sub.get_values_and_class('vec_slc_to_const') == (1, ('global', 'const'))
        assert sub.get_values_and_class('glb_slc_to_const') == (1, ('global', 'const'))

    def test_time_sample_subset(self):
        for time_idx in range(5):
            sub = self.ext.get_subset(3, time_idx)
            sub.check_valid()
            for classes in self.ext.get_valid_classes():
                key = '%s_%s_test' % classes
                if classes[0] == 'time':
                    assert(not classes in sub.get_valid_classes())
                    if classes[1] == 'samples':
                        assert sub.get_values_and_class(key) == (list(range(time_idx, 5 * 7, 5)), ('vector', 'samples'))
                    elif classes[1] == 'slices':
                        assert sub.get_values_and_class(key) == (self.ext.get_values(key), ('vector', 'slices'))
                elif classes[0] == 'vector':
                    if classes[1] == 'samples':
                        assert sub.get_values_and_class(key) == self.ext.get_values_and_class(key)
                    elif classes[1] == 'slices':
                        start = time_idx * 3
                        end = start + 3
                        assert sub.get_values_and_class(key) == (list(range(start, end)), ('vector', 'slices'))
                else:
                    if classes[1] == 'const':
                        assert sub.get_values_and_class(key) == self.ext.get_values_and_class(key)
                    elif classes[1] == 'slices':
                        vals = []
                        for vec_idx in range(7):
                            start = (vec_idx * (3 * 5)) + (time_idx * 3)
                            end = start + 3
                            vals += list(range(start, end))
                        assert sub.get_values_and_class(key) == (vals, ('global', 'slices'))

    def test_time_sample_subset_simplify(self):
        #Test for simplification of time samples that become constant
        vals = []
        for vector_idx in range(7):
            for time_idx in range(5):
                vals.append(time_idx)
        self.ext.get_class_dict(('time', 'samples'))['time_smp_to_const'] = \
            vals

        for time_idx in range(5):
            sub = self.ext.get_subset(3, time_idx)
            sub.check_valid()
            assert sub.get_values_and_class('time_smp_to_const') == (time_idx, ('global', 'const'))

        #Test for simplification of vector slices that become constant
        vals = []
        for time_idx in range(5):
            for slice_idx in range(3):
                if time_idx == 1:
                    vals.append(1)
                else:
                    vals.append(slice_idx)
        self.ext.get_class_dict(('vector', 'slices'))['vec_slc_to_const'] = \
            vals

        #Test simplification of global slices that become constant
        vals = []
        for vector_idx in range(7):
            for time_idx in range(5):
                for slice_idx in range(3):
                    if time_idx == 1:
                        vals.append(1)
                    else:
                        vals.append(vector_idx)
        self.ext.get_class_dict(('global', 'slices'))['glb_slc_to_const'] = \
            vals

        sub = self.ext.get_subset(3, 1)
        sub.check_valid()
        assert sub.get_values_and_class('vec_slc_to_const') == (1, ('global', 'const'))
        assert sub.get_values_and_class('glb_slc_to_const') == (1, ('global', 'const'))

        #Test simplification of global slices that become vector slices or
        #samples
        vals = []
        for vector_idx in range(7):
            for time_idx in range(5):
                for slice_idx in range(3):
                    if time_idx == 1:
                        vals.append(slice_idx)
                    else:
                        vals.append(vector_idx)
        self.ext.get_class_dict(('global', 'slices'))['glb_slc_to_vec'] = \
            vals

        for time_idx in range(5):
            sub = self.ext.get_subset(3, time_idx)
            sub.check_valid()
            if time_idx == 1:
                assert sub.get_values_and_class('glb_slc_to_vec') == (list(range(3)), ('vector', 'slices'))
            else:
                assert sub.get_values_and_class('glb_slc_to_vec') == (list(range(7)), ('vector', 'samples'))

    def test_vector_sample_subset(self):
        for vector_idx in range(7):
            sub = self.ext.get_subset(4, vector_idx)
            sub.check_valid()
            for classes in self.ext.get_valid_classes():
                key = '%s_%s_test' % classes
                if classes[0] == 'vector':
                    assert(not classes in sub.get_valid_classes())
                    if classes[1] == 'samples':
                        assert sub.get_values_and_class(key) == (vector_idx, ('global', 'const'))
                    elif classes[1] == 'slices':
                        assert sub.get_values_and_class(key) == (list(range(3 * 5)), ('global', 'slices'))
                elif classes[0] == 'time':
                    if classes[1] == 'samples': #Could be const
                        start = vector_idx * 5
                        end = start + 5
                        assert sub.get_values_and_class(key) == (list(range(start, end)), classes)
                    elif classes[1] == 'slices':
                        assert sub.get_values_and_class(key) == self.ext.get_values_and_class(key)
                else:
                    if classes[1] == 'const':
                        assert sub.get_values_and_class(key) == self.ext.get_values_and_class(key)
                    elif classes[1] == 'slices': #Could be const or time samples or time slices
                        start = vector_idx * (3 * 5)
                        end = start + (3 * 5)
                        assert sub.get_values_and_class(key) == (list(range(start, end)), classes)

    def test_vector_sample_subset_simplify(self):

        #Test for simplification of time samples that become constant
        vals = []
        for vector_idx in range(7):
            for time_idx in range(5):
                if vector_idx == 1:
                    vals.append(1)
                else:
                    vals.append(time_idx)
        self.ext.get_class_dict(('time', 'samples'))['time_smp_to_const'] = \
            vals
        sub = self.ext.get_subset(4, 1)
        assert sub.get_values_and_class('time_smp_to_const') == (1, ('global', 'const'))

        #Test for simplification of global slices that become constant, time
        #samples, or time slices
        vals = []
        for vector_idx in range(7):
            for time_idx in range(5):
                for slice_idx in range(3):
                    if vector_idx == 1:
                        vals.append(1)
                    elif vector_idx == 2:
                        vals.append(slice_idx)
                    else:
                        vals.append(time_idx)
        self.ext.get_class_dict(('global', 'slices'))['glb_slc'] = \
            vals

        for vector_idx in range(7):
            sub = self.ext.get_subset(4, vector_idx)
            sub.check_valid()
            if vector_idx == 1:
                assert sub.get_values_and_class('glb_slc') == (1, ('global', 'const'))
            elif vector_idx == 2:
                assert sub.get_values_and_class('glb_slc') == (list(range(3)), ('time', 'slices'))
            else:
                assert sub.get_values_and_class('glb_slc') == (list(range(5)), ('time', 'samples'))

class TestChangeClass(object):
    def setUp(self):
        self.ext = dcmmeta.DcmMetaExtension.make_empty((2, 2, 3, 5, 7),
                                                       np.eye(4),
                                                       np.eye(4),
                                                       2
                                                      )
        #Add an element to every classification
        for classes in self.ext.get_valid_classes():
            key = '%s_%s_test' % classes
            mult = self.ext.get_multiplicity(classes)
            if mult == 1:
                vals = 0
            else:
                vals = list(range(mult))
            self.ext.get_class_dict(classes)[key] = vals

    def setup_method(self, setUp):
        self.setUp()

    def test_change_none(self):
        assert self.ext._get_changed_class('None_test', ('global', 'const')) == None
        assert self.ext._get_changed_class('None_test', ('global', 'slices')) == [None] * (3 * 5 * 7)
        assert self.ext._get_changed_class('None_test', ('time', 'samples')) ==[None] * (5 * 7)
        assert self.ext._get_changed_class('None_test', ('time', 'slices')) == [None] * 3
        assert self.ext._get_changed_class('None_test', ('vector', 'samples')) == [None] * 7
        assert self.ext._get_changed_class('None_test', ('vector', 'slices')) == [None] * (3 * 5)

    def test_change_global_const(self):
        assert self.ext._get_changed_class('global_const_test', ('global', 'slices')) == [0] * (3 * 5 * 7)
        assert self.ext._get_changed_class('global_const_test', ('time', 'samples')) == [0] * (5 * 7)
        assert self.ext._get_changed_class('global_const_test', ('time', 'slices')) == [0] * 3
        assert self.ext._get_changed_class('global_const_test', ('vector', 'samples')) == [0] * 7
        assert self.ext._get_changed_class('global_const_test', ('vector', 'slices')) == [0] * (3 * 5)

    def test_change_vector_samples(self):
        vals = []
        for vector_idx in range(7):
            vals += [vector_idx] * 15
        assert self.ext._get_changed_class('vector_samples_test', ('global', 'slices')) == vals
        vals = []
        for vector_idx in range(7):
            vals += [vector_idx] * 5
        assert self.ext._get_changed_class('vector_samples_test', ('time', 'samples')) == vals

    def test_change_time_samples(self):
        vals = []
        for time_idx in range(5 * 7):
            vals += [time_idx] * 3
        assert self.ext._get_changed_class('time_samples_test', ('global', 'slices')) == vals

    def test_time_slices(self):
        vals = []
        for time_idx in range(5 * 7):
            vals += list(range(3))
        assert self.ext._get_changed_class('time_slices_test', ('global', 'slices')) == vals
        vals = []
        for time_idx in range(5):
            vals += list(range(3))
        assert self.ext._get_changed_class('time_slices_test', ('vector', 'slices')) == vals

    def test_vector_slices(self):
        vals = []
        for vector_idx in range(7):
            vals += list(range(15))
        assert self.ext._get_changed_class('vector_slices_test', ('global', 'slices')) == vals


def test_from_sequence_2d_to_3d():
    ext1 = dcmmeta.DcmMetaExtension.make_empty((2, 2, 1),
                                               np.eye(4),
                                               np.eye(4),
                                               2)
    ext1.get_class_dict(('global', 'const'))['const'] = 1
    ext1.get_class_dict(('global', 'const'))['var'] = 1
    ext1.get_class_dict(('global', 'const'))['missing'] = 1
    ext2 = dcmmeta.DcmMetaExtension.make_empty((2, 2, 1),
                                               np.eye(4),
                                               np.eye(4),
                                               2)
    ext2.get_class_dict(('global', 'const'))['const'] = 1
    ext2.get_class_dict(('global', 'const'))['var'] = 2

    merged = dcmmeta.DcmMetaExtension.from_sequence([ext1, ext2], 2)
    assert merged.get_values_and_class('const') == (1, ('global', 'const'))
    assert merged.get_values_and_class('var') == ([1, 2], ('global', 'slices'))
    assert merged.get_values_and_class('missing') == ([1, None], ('global', 'slices'))

def test_from_sequence_3d_to_4d():
    for dim_name, dim in (('time', 3), ('vector', 4)):
        ext1 = dcmmeta.DcmMetaExtension.make_empty((2, 2, 2),
                                                   np.eye(4),
                                                   np.eye(4),
                                                   2)
        ext1.get_class_dict(('global', 'const'))['global_const_const'] = 1
        ext1.get_class_dict(('global', 'const'))['global_const_var'] = 1
        ext1.get_class_dict(('global', 'const'))['global_const_missing'] = 1
        ext1.get_class_dict(('global', 'slices'))['global_slices_const'] = [0, 1]
        ext1.get_class_dict(('global', 'slices'))['global_slices_var'] = [0, 1]
        ext1.get_class_dict(('global', 'slices'))['global_slices_missing'] = [0, 1]
        ext2 = dcmmeta.DcmMetaExtension.make_empty((2, 2, 2),
                                                   np.eye(4),
                                                   np.eye(4),
                                                   2)
        ext2.get_class_dict(('global', 'const'))['global_const_const'] = 1
        ext2.get_class_dict(('global', 'const'))['global_const_var'] = 2
        ext2.get_class_dict(('global', 'slices'))['global_slices_const'] = [0, 1]
        ext2.get_class_dict(('global', 'slices'))['global_slices_var'] = [1, 2]

        merged = dcmmeta.DcmMetaExtension.from_sequence([ext1, ext2], dim)
        assert merged.get_values_and_class('global_const_const'), (1, ('global', 'const'))
        assert merged.get_values_and_class('global_const_var') == ([1, 2], (dim_name, 'samples'))
        assert merged.get_values_and_class('global_const_missing') == ([1, None], (dim_name, 'samples'))
        assert merged.get_values_and_class('global_slices_const') == ([0, 1], (dim_name, 'slices'))
        assert merged.get_values_and_class('global_slices_var') == ([0, 1, 1, 2], ('global', 'slices'))
        assert merged.get_values_and_class('global_slices_missing') == ([0, 1, None, None], ('global', 'slices'))

def test_from_sequence_4d_time_to_5d():
    ext1 = dcmmeta.DcmMetaExtension.make_empty((2, 2, 2, 2),
                                               np.eye(4),
                                               np.eye(4),
                                               2)
    ext1.get_class_dict(('global', 'const'))['global_const_const'] = 1
    ext1.get_class_dict(('global', 'const'))['global_const_var'] = 1
    ext1.get_class_dict(('global', 'const'))['global_const_missing'] = 1
    ext1.get_class_dict(('global', 'slices'))['global_slices_const'] = [0, 1, 2, 3]
    ext1.get_class_dict(('global', 'slices'))['global_slices_var'] = [0, 1, 2, 3]
    ext1.get_class_dict(('global', 'slices'))['global_slices_missing'] = [0, 1, 2, 3]
    ext1.get_class_dict(('time', 'samples'))['time_samples_const'] = [0, 1]
    ext1.get_class_dict(('time', 'samples'))['time_samples_var'] = [0, 1]
    ext1.get_class_dict(('time', 'samples'))['time_samples_missing'] = [0, 1]
    ext1.get_class_dict(('time', 'slices'))['time_slices_const'] = [0, 1]
    ext1.get_class_dict(('time', 'slices'))['time_slices_var'] = [0, 1]
    ext1.get_class_dict(('time', 'slices'))['time_slices_missing'] = [0, 1]

    ext2 = dcmmeta.DcmMetaExtension.make_empty((2, 2, 2, 2),
                                               np.eye(4),
                                               np.eye(4),
                                               2)
    ext2.get_class_dict(('global', 'const'))['global_const_const'] = 1
    ext2.get_class_dict(('global', 'const'))['global_const_var'] = 2
    ext2.get_class_dict(('global', 'slices'))['global_slices_const'] = [0, 1, 2, 3]
    ext2.get_class_dict(('global', 'slices'))['global_slices_var'] = [1, 2, 3, 4]
    ext2.get_class_dict(('time', 'samples'))['time_samples_const'] = [0, 1]
    ext2.get_class_dict(('time', 'samples'))['time_samples_var'] = [1, 2]
    ext2.get_class_dict(('time', 'slices'))['time_slices_const'] = [0, 1]
    ext2.get_class_dict(('time', 'slices'))['time_slices_var'] = [1, 2]

    merged = dcmmeta.DcmMetaExtension.from_sequence([ext1, ext2], 4)
    assert merged.get_values_and_class('global_const_const') == (1, ('global', 'const'))
    assert merged.get_values_and_class('global_const_var') == ([1, 2], ('vector', 'samples'))
    assert merged.get_values_and_class('global_const_missing') == ([1, None], ('vector', 'samples'))
    assert merged.get_values_and_class('global_slices_const') == ([0, 1, 2, 3], ('vector', 'slices'))
    assert merged.get_values_and_class('global_slices_var') == ([0, 1, 2, 3, 1, 2, 3, 4], ('global', 'slices'))
    assert merged.get_values_and_class('global_slices_missing') == ([0, 1, 2, 3, None, None, None, None], ('global', 'slices'))
    assert merged.get_values_and_class('time_samples_const') == ([0, 1, 0, 1], ('time', 'samples'))
    assert merged.get_values_and_class('time_samples_var') == ([0, 1, 1, 2], ('time', 'samples'))
    assert merged.get_values_and_class('time_samples_missing') == ([0, 1, None, None], ('time', 'samples'))
    assert merged.get_values_and_class('time_slices_const') == ([0, 1], ('time', 'slices'))
    assert merged.get_values_and_class('time_slices_var') == ([0, 1, 0, 1, 1, 2, 1, 2], ('global', 'slices'))
    assert merged.get_values_and_class('time_slices_missing') == ([0, 1, 0, 1, None, None, None, None], ('global', 'slices'))

def test_from_sequence_no_slc():
    ext1 = dcmmeta.DcmMetaExtension.make_empty((2, 2, 2), np.eye(4))
    ext2 = dcmmeta.DcmMetaExtension.make_empty((2, 2, 2), np.eye(4))
    merged = dcmmeta.DcmMetaExtension.from_sequence([ext1, ext2], 4)

def test_nifti_wrapper_init():
    nii = nb.Nifti1Image(np.zeros((5, 5, 5)), np.eye(4))
    with pytest.raises(nifti_wrapper.MissingExtensionError):
        nifti_wrapper.NiftiWrapper(nii)
    hdr = nii.header
    ext = dcmmeta.DcmMetaExtension.make_empty((5, 5, 5), np.eye(4))
    hdr.extensions.append(ext)
    nw = nifti_wrapper.NiftiWrapper(nii)
    assert nw.meta_ext == ext

    nii2 = nb.Nifti1Image(np.zeros((5, 5, 5)), np.eye(4))
    nw2 = nifti_wrapper.NiftiWrapper(nii, True)
    ext2 = nw2.meta_ext
    assert ext == ext2

class TestMetaValid(object):
    def setUp(self):
        nii = nb.Nifti1Image(np.zeros((5, 5, 5, 7, 9)), np.eye(4))
        hdr = nii.header
        hdr.set_dim_info(None, None, 2)
        self.nw = nifti_wrapper.NiftiWrapper(nii, True)
        for classes in self.nw.meta_ext.get_valid_classes():
            assert(self.nw.meta_valid(classes))

    def setup_method(self, setUp):
        self.setUp()

    def test_time_samples_changed(self):
        self.nw.meta_ext.shape = (5, 5, 5, 3, 9)
        for classes in self.nw.meta_ext.get_valid_classes():
            if classes in (('time', 'samples'),
                           ('vector', 'slices'),
                           ('global', 'slices')):
                assert self.nw.meta_valid(classes) == False
            else:
                assert(self.nw.meta_valid(classes))

    def test_vector_samples_changed(self):
        self.nw.meta_ext.shape = (5, 5, 5, 7, 3)
        for classes in self.nw.meta_ext.get_valid_classes():
            if classes in (('time', 'samples'),
                           ('vector', 'samples'),
                           ('global', 'slices')):
                assert self.nw.meta_valid(classes) == False
            else:
                assert(self.nw.meta_valid(classes))

    def test_slice_dim_changed(self):
        self.nw.meta_ext.slice_dim = 0
        for classes in self.nw.meta_ext.get_valid_classes():
            if classes[1] == 'slices':
                assert self.nw.meta_valid(classes) == False
            else:
                assert(self.nw.meta_valid(classes))

    def test_slice_dir_changed(self):
        aff = self.nw.nii_img.affine
        aff[:] = np.c_[aff[2, :],
                       aff[1, :],
                       aff[0, :],
                       aff[3, :],
                      ]

        for classes in self.nw.meta_ext.get_valid_classes():
            if classes[1] == 'slices':
                assert self.nw.meta_valid(classes) == False
            else:
                assert(self.nw.meta_valid(classes))

class TestGetMeta(object):
    def setUp(self):
        nii = nb.Nifti1Image(np.zeros((5, 5, 5, 7, 9)), np.eye(4))
        hdr = nii.header
        hdr.set_dim_info(None, None, 2)
        self.nw = nifti_wrapper.NiftiWrapper(nii, True)

        #Add an element to every classification
        for classes in self.nw.meta_ext.get_valid_classes():
            key = '%s_%s_test' % classes
            mult = self.nw.meta_ext.get_multiplicity(classes)
            if mult == 1:
                vals = 0
            else:
                vals = list(range(mult))
            self.nw.meta_ext.get_class_dict(classes)[key] = vals

    def setup_method(self, setUp):
        self.setUp()

    def test_get_item(self):
        for classes in self.nw.meta_ext.get_valid_classes():
            key = '%s_%s_test' % classes
            if classes == ('global', 'const'):
                assert self.nw[key] == 0
            else:
                with pytest.raises(KeyError):
                    self.nw.__getitem__(key)

    def test_get_const(self):
        assert self.nw.get_meta('global_const_test') == 0
        assert self.nw.get_meta('global_const_test', (0, 0, 0, 0, 0)) == 0
        assert self.nw.get_meta('global_const_test', (0, 0, 3, 4, 5)) == 0

    def test_get_global_slices(self):
        assert self.nw.get_meta('global_slices_test') == None
        assert self.nw.get_meta('global_slices_test', None, 0) == 0
        for vector_idx in range(9):
            for time_idx in range(7):
                for slice_idx in range(5):
                    idx = (0, 0, slice_idx, time_idx, vector_idx)
                    assert self.nw.get_meta('global_slices_test', idx) == slice_idx + (time_idx * 5) + (vector_idx * 7 * 5)

    def test_get_vector_slices(self):
        assert self.nw.get_meta('vector_slices_test') == None
        assert self.nw.get_meta('vector_slices_test', None, 0) == 0
        for vector_idx in range(9):
            for time_idx in range(7):
                for slice_idx in range(5):
                    idx = (0, 0, slice_idx, time_idx, vector_idx)
                    assert self.nw.get_meta('vector_slices_test', idx) == slice_idx + (time_idx * 5)

    def test_get_time_slices(self):
        assert self.nw.get_meta('time_slices_test') == None
        assert self.nw.get_meta('time_slices_test', None, 0) == 0
        for vector_idx in range(9):
            for time_idx in range(7):
                for slice_idx in range(5):
                    idx = (0, 0, slice_idx, time_idx, vector_idx)
                    assert self.nw.get_meta('time_slices_test', idx) == slice_idx

    def test_get_vector_samples(self):
        assert self.nw.get_meta('vector_samples_test') == None
        assert self.nw.get_meta('vector_samples_test', None, 0) == 0
        for vector_idx in range(9):
            for time_idx in range(7):
                for slice_idx in range(5):
                    idx = (0, 0, slice_idx, time_idx, vector_idx)
                    assert self.nw.get_meta('vector_samples_test', idx) == vector_idx

    def get_time_samples(self):
        assert self.nw.get_meta('time_samples_test') == None
        assert self.nw.get_meta('time_samples_test', None, 0) == 0
        for vector_idx in range(9):
            for time_idx in range(7):
                for slice_idx in range(5):
                    idx = (0, 0, slice_idx, time_idx, vector_idx)
                    assert self.nw.get_meta('time_samples_test', idx) == time_idx + (vector_idx * 7)

class TestSplit(object):
    def setUp(self):
        self.arr = np.arange(3 * 3 * 3 * 5 * 7, dtype=np.int32).reshape(3, 3, 3, 5, 7)
        nii = nb.Nifti1Image(self.arr, np.diag([1.1, 1.1, 1.1, 1.0]))
        hdr = nii.header
        hdr.set_dim_info(None, None, 2)
        self.nw = nifti_wrapper.NiftiWrapper(nii, True)

        #Add an element to every classification
        for classes in self.nw.meta_ext.get_valid_classes():
            key = '%s_%s_test' % classes
            mult = self.nw.meta_ext.get_multiplicity(classes)
            if mult == 1:
                vals = 0
            else:
                vals = list(range(mult))
            self.nw.meta_ext.get_class_dict(classes)[key] = vals

    def setup_method(self, setUp):
        self.setUp()

    def test_split_slice(self):
        for split_idx, nw_split in enumerate(self.nw.split(2)):
            assert nw_split.nii_img.shape == (3, 3, 1, 5, 7)
            assert(np.allclose(nw_split.nii_img.affine,
                            np.c_[[1.1, 0.0, 0.0, 0.0],
                                  [0.0, 1.1, 0.0, 0.0],
                                  [0.0, 0.0, 1.1, 0.0],
                                  [0.0, 0.0, 1.1*split_idx, 1.0]
                                 ]
                           )
               )
            assert(np.all(np.asanyarray(nw_split.nii_img.dataobj) ==
                       self.arr[:, :, split_idx:split_idx+1, :, :])
               )

    def test_split_time(self):
        for split_idx, nw_split in enumerate(self.nw.split(3)):
            assert nw_split.nii_img.shape == (3, 3, 3, 1, 7)
            assert(np.allclose(nw_split.nii_img.affine,
                            np.diag([1.1, 1.1, 1.1, 1.0])))
            assert(np.all(np.asanyarray(nw_split.nii_img.dataobj) ==
                       self.arr[:, :, :, split_idx:split_idx+1, :])
               )

    def test_split_vector(self):
        for split_idx, nw_split in enumerate(self.nw.split(4)):
            assert nw_split.nii_img.shape == (3, 3, 3, 5)
            assert(np.allclose(nw_split.nii_img.affine,
                            np.diag([1.1, 1.1, 1.1, 1.0])))
            assert(np.all(np.asanyarray(nw_split.nii_img.dataobj) ==
                       self.arr[:, :, :, :, split_idx])
               )

def test_split_keep_spatial():
    arr = np.arange(3 * 3 * 3, dtype=np.int32).reshape(3, 3, 3)
    nii = nb.Nifti1Image(arr, np.eye(4))
    nw = nifti_wrapper.NiftiWrapper(nii, True)

    for split_idx, nw_split in enumerate(nw.split(2)):
        assert nw_split.nii_img.shape == (3, 3, 1)


def test_from_dicom():
    data_dir = path.join(test_dir,
                         'data',
                         'dcmstack',
                         '2D_16Echo_qT2')
    src_fn = path.join(data_dir, 'TE_40_SlcPos_-33.707626341697.dcm')
    src_dcm = pydicom.dcmread(src_fn)
    src_dw = nb.nicom.dicomwrappers.wrapper_from_data(src_dcm)
    meta = {'EchoTime': 40}
    nw = nifti_wrapper.NiftiWrapper.from_dicom(src_dcm, meta)
    hdr = nw.nii_img.header
    assert nw.nii_img.shape == (192, 192, 1)
    assert(np.allclose(np.dot(np.diag([-1., -1., 1., 1.]), src_dw.affine),
                    nw.nii_img.affine)
       )
    assert hdr.get_xyzt_units() == ('mm', 'sec')
    assert hdr.get_dim_info() == (0, 1, 2)
    assert nw.meta_ext.get_values_and_class('EchoTime') == (40, ('global', 'const'))

def test_from_2d_slice_to_3d():
    slice_nws = []
    for idx in range(3):
        arr = np.arange(idx * (4 * 4), (idx + 1) * (4 * 4), dtype=np.int32).reshape(4, 4, 1)
        aff = np.diag((1.1, 1.1, 1.1, 1.0))
        aff[:3, 3] += [0.0, 0.0, idx * 0.5]
        nii = nb.Nifti1Image(arr, aff)
        hdr = nii.header
        hdr.set_dim_info(0, 1, 2)
        hdr.set_xyzt_units('mm', 'sec')
        nw = nifti_wrapper.NiftiWrapper(nii, True)
        nw.meta_ext.get_class_dict(('global', 'const'))['EchoTime'] = 40
        nw.meta_ext.get_class_dict(('global', 'const'))['SliceLocation'] = idx
        slice_nws.append(nw)

    merged = nifti_wrapper.NiftiWrapper.from_sequence(slice_nws, 2)
    assert merged.nii_img.shape == (4, 4, 3)
    assert(np.allclose(merged.nii_img.affine,
                    np.diag((1.1, 1.1, 0.5, 1.0)))
       )
    assert merged.meta_ext.get_values_and_class('EchoTime') == (40, ('global', 'const'))
    assert merged.meta_ext.get_values_and_class('SliceLocation') == (list(range(3)), ('global', 'slices'))
    merged_hdr = merged.nii_img.header
    assert merged_hdr.get_dim_info() == (0, 1, 2)
    assert merged_hdr.get_xyzt_units() == ('mm', 'sec')
    merged_data = np.asanyarray(merged.nii_img.dataobj)
    for idx in range(3):
        assert(np.all(merged_data[:, :, idx] ==
                   np.arange(idx * (4 * 4), (idx + 1) * (4 * 4), dtype=np.int32).reshape(4, 4))
           )

def test_from_3d_time_to_4d():
    time_nws = []
    for idx in range(3):
        arr = np.arange(idx * (4 * 4 * 4),
                        (idx + 1) * (4 * 4 * 4), 
                        dtype=np.int32
                       ).reshape(4, 4, 4)
        nii = nb.Nifti1Image(arr, np.diag((1.1, 1.1, 1.1, 1.0)))
        hdr = nii.header
        hdr.set_dim_info(0, 1, 2)
        hdr.set_xyzt_units('mm', 'sec')
        nw = nifti_wrapper.NiftiWrapper(nii, True)
        const_meta = nw.meta_ext.get_class_dict(('global', 'const'))
        const_meta['PatientID'] = 'Test'
        const_meta['EchoTime'] = idx
        glb_slice_meta = nw.meta_ext.get_class_dict(('global', 'slices'))
        glb_slice_meta['SliceLocation'] = list(range(4))
        glb_slice_meta['AcquisitionTime'] = list(range(idx * 4, (idx + 1) * 4))
        time_nws.append(nw)

    merged = nifti_wrapper.NiftiWrapper.from_sequence(time_nws, 3)
    assert merged.nii_img.shape == (4, 4, 4, 3)
    assert(np.allclose(merged.nii_img.affine,
                    np.diag((1.1, 1.1, 1.1, 1.0)))
       )
    assert merged.meta_ext.get_values_and_class('PatientID') == ('Test', ('global', 'const'))
    assert merged.meta_ext.get_values_and_class('EchoTime') == ([0, 1, 2], ('time', 'samples'))
    assert merged.meta_ext.get_values_and_class('SliceLocation') == (list(range(4)), ('time', 'slices'))
    assert merged.meta_ext.get_values_and_class('AcquisitionTime') == (list(range(4 * 3)), ('global', 'slices'))
    merged_hdr = merged.nii_img.header
    assert merged_hdr.get_dim_info() == (0, 1, 2)
    assert merged_hdr.get_xyzt_units() == ('mm', 'sec')
    merged_data = np.asanyarray(merged.nii_img.dataobj)
    for idx in range(3):
        assert(np.all(merged_data[:, :, :, idx] ==
                   np.arange(idx * (4 * 4 * 4),
                             (idx + 1) * (4 * 4 * 4), 
                             dtype=np.int32).reshape(4, 4, 4))
           )

def test_from_3d_vector_to_4d():
    vector_nws = []
    for idx in range(3):
        arr = np.arange(idx * (4 * 4 * 4),
                        (idx + 1) * (4 * 4 * 4), 
                        dtype=np.int32
                       ).reshape(4, 4, 4)
        nii = nb.Nifti1Image(arr, np.diag((1.1, 1.1, 1.1, 1.0)))
        hdr = nii.header
        hdr.set_dim_info(0, 1, 2)
        hdr.set_xyzt_units('mm', 'sec')
        nw = nifti_wrapper.NiftiWrapper(nii, True)
        const_meta = nw.meta_ext.get_class_dict(('global', 'const'))
        const_meta['PatientID'] = 'Test'
        const_meta['EchoTime'] = idx
        glb_slice_meta = nw.meta_ext.get_class_dict(('global', 'slices'))
        glb_slice_meta['SliceLocation'] = list(range(4))
        glb_slice_meta['AcquisitionTime'] = list(range(idx * 4, (idx + 1) * 4))
        vector_nws.append(nw)

    merged = nifti_wrapper.NiftiWrapper.from_sequence(vector_nws, 4)
    assert merged.nii_img.shape == (4, 4, 4, 1, 3)
    assert(np.allclose(merged.nii_img.affine,
                    np.diag((1.1, 1.1, 1.1, 1.0)))
       )
    assert merged.meta_ext.get_values_and_class('PatientID') == ('Test', ('global', 'const'))
    assert merged.meta_ext.get_values_and_class('EchoTime') == ([0, 1, 2], ('vector', 'samples'))
    assert merged.meta_ext.get_values_and_class('SliceLocation') == (list(range(4)), ('vector', 'slices'))
    assert merged.meta_ext.get_values_and_class('AcquisitionTime') == (list(range(4 * 3)), ('global', 'slices'))
    merged_hdr = merged.nii_img.header
    assert merged_hdr.get_dim_info() == (0, 1, 2)
    assert merged_hdr.get_xyzt_units(), ('mm', 'sec')
    merged_data = np.asanyarray(merged.nii_img.dataobj)
    for idx in range(3):
        assert(np.all(merged_data[:, :, :, 0, idx] ==
                   np.arange(idx * (4 * 4 * 4),
                             (idx + 1) * (4 * 4 * 4), 
                             dtype=np.int32).reshape(4, 4, 4))
           )

def test_merge_inconsistent_hdr():
    #Test that inconsistent header data does not make it into the merged
    #result
    time_nws = []
    for idx in range(3):
        arr = np.arange(idx * (4 * 4 * 4),
                        (idx + 1) * (4 * 4 * 4), 
                        dtype=np.int32
                       ).reshape(4, 4, 4)
        nii = nb.Nifti1Image(arr, np.diag((1.1, 1.1, 1.1, 1.0)))
        hdr = nii.header
        if idx == 1:
            hdr.set_dim_info(1, 0, 2)
            hdr.set_xyzt_units('mm', None)
        else:
            hdr.set_dim_info(0, 1, 2)
            hdr.set_xyzt_units('mm', 'sec')
        nw = nifti_wrapper.NiftiWrapper(nii, True)
        time_nws.append(nw)

    merged = nifti_wrapper.NiftiWrapper.from_sequence(time_nws)
    merged_hdr = merged.nii_img.header
    assert merged_hdr.get_dim_info() == (None, None, 2)
    assert merged_hdr.get_xyzt_units(), ('mm', 'unknown')

def test_merge_with_slc_and_without():
    #Test merging two data sets where one has per slice meta and other does not
    input_nws = []
    for idx in range(3):
        arr = np.arange(idx * (4 * 4 * 4),
                        (idx + 1) * (4 * 4 * 4), 
                        dtype=np.int32
                       ).reshape(4, 4, 4)
        nii = nb.Nifti1Image(arr, np.diag((1.1, 1.1, 1.1, 1.0)))
        hdr = nii.header
        if idx == 0:
            hdr.set_dim_info(0, 1, 2)
        hdr.set_xyzt_units('mm', 'sec')
        nw = nifti_wrapper.NiftiWrapper(nii, True)
        const_meta = nw.meta_ext.get_class_dict(('global', 'const'))
        const_meta['PatientID'] = 'Test'
        const_meta['EchoTime'] = idx
        if idx == 0:
            glb_slice_meta = nw.meta_ext.get_class_dict(('global', 'slices'))
            glb_slice_meta['SliceLocation'] = list(range(4))
        input_nws.append(nw)

    merged = nifti_wrapper.NiftiWrapper.from_sequence(input_nws)
    assert merged.nii_img.shape == (4, 4, 4, 3)
    assert(np.allclose(merged.nii_img.affine,
                    np.diag((1.1, 1.1, 1.1, 1.0)))
       )
    assert merged.meta_ext.get_values_and_class('PatientID') == ('Test', ('global', 'const'))
    assert merged.meta_ext.get_values_and_class('EchoTime') == ([0, 1, 2], ('time', 'samples'))
    assert merged.meta_ext.get_values_and_class('SliceLocation') == (list(range(4)) + ([None] * 8), ('global', 'slices'))
    merged_hdr = merged.nii_img.header
    assert merged_hdr.get_xyzt_units() == ('mm', 'sec')



from setuptools import setup, find_packages
import sys, os

# Most of the relevant info is stored in this file
info_file = os.path.join('src', 'dcmstack', 'info.py')
exec(open(info_file).read())


setup(name=NAME,
      python_requires=">=3.8",
      description=DESCRIPTION,
      long_description=open("README.rst").read(),
      long_description_content_type="text/x-rst",
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      maintainer=MAINTAINER,
      maintainer_email=MAINTAINER_EMAIL,
      url=URL,
      classifiers=CLASSIFIERS,
      platforms=PLATFORMS,
      version=VERSION,
      provides=PROVIDES,
      packages=find_packages('src'),
      package_dir = {'':'src'},
      install_requires=INSTALL_REQUIRES,
      extras_require=EXTRAS_REQUIRES,
      entry_points = {'console_scripts' : \
                          ['dcmstack = dcmstack.dcmstack_cli:main',
                           'nitool = dcmstack.nitool_cli:main',
                          ],
                     },
      test_suite = 'pytest'
     )



"""
Tests for dcmstack.extract
"""
import sys, warnings
from os import path

import pytest
from nibabel.nicom import csareader
try:
    import pydicom
except ImportError:
    import dicom as pydicom

from . import test_dir, src_dir

with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    from dcmstack import extract

class TestCsa(object):
    def setup_method(self, method):
        data_fn = path.join(test_dir, 'data', 'extract', 'csa_test.dcm')
        self.data = pydicom.dcmread(data_fn)

    def teardown_method(self, method):
        del self.data

    def test_simplify(self):
        assert extract.simplify_csa_dict(None) == None
        csa_dict = csareader.read(self.data[pydicom.tag.Tag(0x29, 0x1010)].value)
        simp_dict = extract.simplify_csa_dict(csa_dict)
        for tag in csa_dict['tags']:
            items = csa_dict['tags'][tag]['items']
            if len(items) == 0:
                assert(not tag in simp_dict)
            elif len(items) == 1:
                assert simp_dict[tag] == items[0]
            else:
                assert simp_dict[tag] == items

    def test_csa_image_trans(self):
        csa_dict = extract.csa_series_trans_func(self.data[(0x29, 0x1010)])
        assert csa_dict["EchoLinePosition"] == 64

    def test_parse_phx_line(self):
        assert(extract._parse_phoenix_line("") is None)
        assert(extract._parse_phoenix_line("#test = 2") is None)

        assert extract._parse_phoenix_line('test = "" 2 ""') == ('test', ' 2 ')
        assert extract._parse_phoenix_line('test = "" #2 ""'), ('test', ' #2 ')
        assert extract._parse_phoenix_line('test = "" 2 ""#=') == ('test', ' 2 ')
        assert extract._parse_phoenix_line("test = 2#") == ('test', 2)
        assert extract._parse_phoenix_line("test = 0x2") == ('test', 2)
        assert extract._parse_phoenix_line("test = 2.") == ('test', 2.0)

        
        with pytest.raises(extract.PhoenixParseError):
            extract._parse_phoenix_line('test = blah')
        with pytest.raises(extract.PhoenixParseError):
            extract._parse_phoenix_line('==')
        with pytest.raises(extract.PhoenixParseError):
            extract._parse_phoenix_line('test')
        with pytest.raises(extract.PhoenixParseError):
            extract._parse_phoenix_line('test = "" 2 ""3')

    def test_csa_series_trans(self):
        csa_dict = extract.csa_series_trans_func(self.data[(0x29, 0x1020)])
        assert csa_dict['MrPhoenixProtocol.sEFISPEC.bEFIDataValid'] == 1

class TestMetaExtractor(object):
    def setup_method(self, method):
        data_fn = path.join(test_dir, 'data', 'extract', 'csa_test.dcm')
        self.data = pydicom.dcmread(data_fn)

    def teardown_method(self,method):
        del self.data

    def test_get_elem_key(self):
        ignore_rules = (extract.ignore_pixel_data,)
        extractor = extract.MetaExtractor(ignore_rules=ignore_rules)
        for key in extractor(self.data):
            assert(key.strip() != '')
            assert(key[0].isalpha())

    def test_get_elem_value(self):
        ignore_rules = (extract.ignore_pixel_data,)
        extractor = extract.MetaExtractor(ignore_rules=ignore_rules)
        for elem in self.data:
            value = extractor._get_elem_value(elem)
            if elem.VM > 1:
                assert(isinstance(value, list))
            if elem.VR in list(extract.unpack_vr_map) + ['DS', 'IS']:
                if elem.VM == 1:
                    assert(not isinstance(value, str))
                else:
                    assert(not any(isinstance(val, str) for val in value))

    def test_dup_trans(self):
        translators = [extract.csa_image_trans, extract.csa_image_trans]
        extractor = extract.MetaExtractor(translators=translators)
        with pytest.raises(ValueError):
            extractor(self.data)

    def test_reloc_private(self):
        extractor = extract.MetaExtractor()
        elem = self.data[(0x29, 0x10)]
        elem.tag = pydicom.tag.Tag((0x29, 0x20))
        self.data[(0x29, 0x20)] = elem
        elem = self.data[(0x29, 0x1010)]
        elem.tag = pydicom.tag.Tag((0x29, 0x2010))
        self.data[(0x29, 0x2010)] = elem
        elem = self.data[(0x29, 0x1020)]
        elem.tag = pydicom.tag.Tag((0x29, 0x2020))
        self.data[(0x29, 0x2020)] = elem
        del self.data[(0x29, 0x10)]
        del self.data[(0x29, 0x1010)]
        del self.data[(0x29, 0x1020)]
        meta_dict = extractor(self.data)
        assert meta_dict["CsaImage.EchoLinePosition"] == 64
        assert meta_dict['CsaSeries.MrPhoenixProtocol.sEFISPEC.bEFIDataValid'] == 1

    def test_non_reloc_private(self):
        extractor = extract.MetaExtractor()
        meta_dict = extractor(self.data)
        assert meta_dict["CsaImage.EchoLinePosition"] == 64
        assert meta_dict['CsaSeries.MrPhoenixProtocol.sEFISPEC.bEFIDataValid'] == 1

    def test_none_vals(self):
        extractor = extract.MetaExtractor()
        self.data.PercentSampling = None
        meta_data = extractor(self.data)
        assert "PercentSampling" not in meta_data
